# -*- coding: iso-8859-1 -*-
# results.py
# Results module
# Copyright 2011-2013 Giuseppe Venturini

# This file is part of the ahkab simulator.
#
# Ahkab is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# Ahkab is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License v2
# along with ahkab.  If not, see <http://www.gnu.org/licenses/>.

"""
This module provides classes for easy, dictionary-like access to simulation
results.

Simulation results are typically returned upon successful simulation of a
circuit and the user is not expected to use their constructor, but rather
to use the methods they provide to access their data set.

Overview of the data interface
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

The solution classes define special methods according to their simulation
type but they all subclass :class:`solution`, which provides the shared
data interface.

The interface allows for accessing the values as::


    >>> ac_sol.keys()
    ['f', 'Vn1', 'Vn2', 'I(V1)', 'I(L1)', 'I(L2)']

Where ``ac_sol`` is a generic example instance of :class:`ac_solution`.

Checking with the ``in`` construct::

    >>> 'Vn1' in ac_sol
    True

Access any variable in the solution object::

    >>> ac_sol['f']
    array([ 6098.38572827,  6102.08394991,  6105.78441425,  6109.48712265,
            6113.19207648,  6116.89927708,  6120.60872583,  6124.32042408,

            [... omissis ...]

            6463.83880528,  6467.75864729,  6471.68086639])

Iterate over the results::

    >>> for var in ac_sol:
    ...     # do something with ac_sol[var]
    ...     pass

Convenience methods are available to identify and access the independent,
swept variable, when it is available::


    >>> ac_sol.get_xlabel()
    'f'
    >>> ac_sol.get_x()
    array([ 6098.38572827,  6102.08394991,  6105.78441425,  6109.48712265,
            6113.19207648,  6116.89927708,  6120.60872583,  6124.32042408,

            [... omissis ...]

            6463.83880528,  6467.75864729,  6471.68086639])

Index of the solution classes
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

.. autosummary::

    ac_solution
    dc_solution
    op_solution
    pss_solution
    pz_solution
    symbolic_solution
    tran_solution

Module reference
~~~~~~~~~~~~~~~~

"""

from __future__ import (unicode_literals, absolute_import,
                        division, print_function)

import sys
import time
import pickle
import re

import numpy as np

from . import circuit
from . import devices
from . import printing
from . import options
from . import constants
from . import csvlib

from .py3compat import text_type
from .__version__ import __version__

csvlib.SEPARATOR = "\t"

class _mutable_data(object):
    def __init__(self):
        self._init_file_done = False
    def _add_data(self, data):
        """Add the data matrix to the results set."""
        csvlib.write_csv(self.filename, data, self.variables, append=self._init_file_done)
        self._init_file_done = True


class solution(object):
    """Base class storing a set of generic simulation results.

    This class is not meant to be accessed directly, rather it is
    subclassed by the classes for the specific simulation solutions.

    **Parameters:**

    circ : circuit instance
        the circuit instance of the simulated circuit.
    outfile : string
        the filename of the save file
    """
    def __init__(self, circ, outfile):
        self.timestamp = time.strftime("%Y-%m-%d %H:%M:%S", time.gmtime())
        self.netlist_file = circ.filename
        self.netlist_title = circ.title
        self.vea = options.vea
        self.ver = options.ver
        self.iea = options.iea
        self.ier = options.ier
        self.gmin = options.gmin
        self.cmin = options.cmin
        self.temp = constants.T
        self.filename = outfile
        self._init_file_done = False

        self.skip_nodes_list = []   # nodi da saltare, solo interni
        self.variables = []
        self.units = case_insensitive_dict()
        self.iter_index = 0
        # Please redefine this sol_type in the subclasses
        self.sol_type = None

    def asarray(self):
        """Return all data.

        .. note::

            This method loads to memory a possibly huge data matrix.

        """
        data, _, _, _ = csvlib.load_csv(self.filename, load_headers=[],
                                        verbose=0)
        return data

    # Access as a dictionary BY VARIABLE NAME:
    def __len__(self):
        """Get the number of variables in the results set."""
        return len(self.variables)

    def __getitem__(self, name):
        """Get a specific variable, as from a dictionary."""
        # data, headers, pos, EOF = csvlib.load_csv(...)
        try:
            data, _, _, _ = csvlib.load_csv(self.filename, load_headers=[name],
                                            nsamples=None, skip=0, verbose=0)
        except ValueError:
            raise KeyError(name)
        return data.reshape((-1,))

    def get(self, name, default=None):
        """Get a solution by variable name."""
        try:
            # data, headers, pos, EOF = csvlib.load_csv(...)
            data, _, _, _ = csvlib.load_csv(self.filename, load_headers=[name],
                                            nsamples=None, skip=0, verbose=0)
        except ValueError:
            return default
        return data.reshape((-1,))

    def has_key(self, name):
        """Determine whether the result set contains a variable."""
        return name.upper() in [v.upper() for v in self.variables]

    def __contains__(self, name):
        """Determine whether the result set contains a variable."""
        return name.upper() in [v.upper() for v in self.variables]

    def keys(self):
        """Get all of the results set's variables names."""
        return self.variables

    def values(self):
        """Get all of the results set's variables values."""
        # data, headers, pos, EOF = csvlib.load_csv(...)
        data, _, _, _ = csvlib.load_csv(self.filename,
                                        load_headers=self.variables,
                                        nsamples=None, skip=0, verbose=0)
        values = [data[i, :] for i in range(data.shape[0])]
        return values

    def items(self):
        return list(zip(self.keys(), self.values()))

    # iterator methods
    def __iter__(self):
        self.iter_index = 0
        self.iter_data, self.iter_headers, _, _ = csvlib.load_csv(self.filename)
        return self

    def next(self):
        return self.__next__()

    def __next__(self):
        if self.iter_index == len(self.iter_headers):
            self.iter_index = 0
            raise StopIteration
        else:
            next_i = self.iter_headers[self.iter_index]
            if hasattr(self.iter_data, 'shape'):
                next_d = self.iter_data[self.iter_index, :]
            else:
                next_d = self.iter_data[self.iter_index]
            nxt = next_i, next_d
            self.iter_index += 1
        return nxt

class op_solution(solution, _mutable_data):
    """OP results

    **Parameters:**

    x : ndarray
        the result set
    error : ndarray
        the residual error after solution,
    circ : circuit instance
        the circuit instance of the simulated circuit
    outfile: str
        the file to write the results to.
        Use "stdout" to write to std output.
    iterations, int, optional
        The number of iterations needed for convergence, if known.
    """
    def __init__(self, x, error, circ, outfile, iterations=0):
        solution.__init__(self, circ, outfile)
        self.sol_type = "OP"
        self.iterations = iterations

        # We have mixed current and voltage results
        # per primi vengono tanti valori di tensioni quanti sono i nodi del circuito meno
        # uno, quindi tante correnti quanti sono gli elementi definiti in tensione presenti
        # (per questo, per misurare una corrente, si può fare uso di generatori di tensione
        # da 0V)

        nv_1 = circ.get_nodes_number() - 1 # numero di soluzioni di tensione (al netto del ref)
        self.results = case_insensitive_dict()
        self.errors = case_insensitive_dict()
        self.x = x

        for index in range(nv_1):
            varname = ("V" + str(circ.nodes_dict[index + 1])).upper()
            self.variables += [varname]
            self.results.update({varname: x[index, 0]})
            self.errors.update({varname: error[index, 0]})
            self.units.update({varname: "V"})
            if circ.is_int_node_internal_only(index+1):
                self.skip_nodes_list.append(index)

        index = nv_1 - 1
        for elem in circ:
            if circuit.is_elem_voltage_defined(elem):
                index = index + 1
                varname = ("I("+elem.part_id.upper()+")").upper()
                self.variables += [varname]
                self.results.update({varname: x[index, 0]})
                self.errors.update({varname: error[index, 0]})
                self.units.update({varname: "A"})

        self._op_keys, self._op_info, self.tot_power = self._get_elements_op(circ, x)

    def __str__(self):
        str_repr = \
            (("OP simulation results for '%s'" % (self.netlist_title,)) +
            ('(netlist %s)'%(self.netlist_file,) if self.netlist_file else '') +
            ('.\nRun on %s, data file %s.\n' % \
             (self.timestamp, self.filename)))
        return str_repr + self.get_table_array()

    def __getitem__(self, name):
        """Get a specific variable, as from a dictionary."""
        if not name.upper() in [v.upper() for v in self.variables]:
            raise KeyError
        his = csvlib.get_headers_index(self.variables, [name], verbose=0)
        return self.x[his]

    def get(self, name, default=None):
        """Get a solution by variable name."""
        try:
            data = self.__getitem__(name)
        except KeyError:
            return default
        return data

    def asarray(self):
        """Get all data as a ``numpy`` array"""
        return self.x

    def get_table_array(self):
        headers = ("Variable", "Units", "Value", "Error", "%")
        table = []
        for v in self.variables:
            if self.results[v] != 0:
                relerror = self.errors[v]/self.results[v]*100.0
            else:
                relerror = 0.0
            line = (v, self.units[v], self.results[v], self.errors[v], '%d' %
                    relerror)
            table.append(line)
        return printing.table(table, headers=headers)

    def _get_elements_op(self, circ, x):
        """Returns"""
        tot_power = 0
        i_index = 0
        nv_1 = circ.get_nodes_number() - 1
        op_info = {}
        op_keys = {}
        for elem in circ:
            ports_v_v = []
            if hasattr(elem, "get_op_info"):
                if elem.is_nonlinear:
                    # build the drive ports vector
                    oports = elem.get_output_ports()
                    for index in range(len(oports)):
                        dports = elem.get_drive_ports(index)
                        ports_v = []
                        for port in dports:
                            tempv = 0
                            if port[0]:
                                tempv = x[port[0]-1]
                            if port[1]:
                                tempv = tempv - x[port[1]-1]
                            ports_v.append(tempv)
                    ports_v_v.append(ports_v)
                else:
                    port = (elem.n1, elem.n2)
                    tempv = 0
                    if port[0]:
                        tempv = x[port[0]-1]
                    if port[1]:
                        tempv = tempv - x[port[1]-1]
                    ports_v_v = ((tempv,),)
                if circuit.is_elem_voltage_defined(elem):
                    i = circ.find_vde_index(elem.part_id)
                    nv_1 = circ.get_nodes_number() - 1
                    opk, opi = elem.get_op_info(ports_v_v, x[nv_1 + i])
                else:
                    opk, opi = elem.get_op_info(ports_v_v)
                if elem.part_id[0].upper() != 'M':
                    if elem.part_id[0].upper() in op_info:
                        op_info.update({elem.part_id[0].upper():
                                        op_info[elem.part_id[0].upper()]+[opi]})
                        #assert set(opk) == set(op_keys[elem.part_id[0].upper()])
                    else:
                        op_info.update({elem.part_id[0].upper():[opi]})
                        op_keys.update({elem.part_id[0].upper():[opk]})
                else:
                    op_info.update({elem.part_id.upper():opi})
                    op_keys.update({elem.part_id.upper():[[]]})

            if isinstance(elem, devices.GISource):
                v = 0
                v = v + x[elem.n1-1] if elem.n1 != 0 else v
                v = v - x[elem.n2-1] if elem.n2 != 0 else v
                vs = 0
                vs = vs + x[elem.n1-1] if elem.sn1 != 0 else vs
                vs = vs - x[elem.n2-1] if elem.sn2 != 0 else vs
                tot_power = tot_power - v*vs*elem.alpha
            elif isinstance(elem, devices.ISource):
                v = 0
                v = v + x[elem.n1-1] if elem.n1 != 0 else v
                v = v - x[elem.n2-1] if elem.n2 != 0 else v
                tot_power = tot_power - v*elem.I()
            elif isinstance(elem, devices.VSource) or \
                 isinstance(elem, devices.EVSource):
                v = 0
                v = v + x[elem.n1-1] if elem.n1 != 0 else v
                v = v - x[elem.n2-1] if elem.n2 != 0 else v
                tot_power = tot_power - v*x[nv_1 + i_index, 0]
                i_index = i_index + 1
            elif isinstance(elem, devices.FISource):
                local_i_index = 0
                found_source = False
                for e in circ:
                    if circuit.is_elem_voltage_defined(e):
                        if isinstance(e, devices.VSource) and e.part_id.lower() == elem.source_id.lower():
                            found_source = True
                            break
                        else:
                            local_i_index += 1
                if not found_source:
                    raise RuntimeError("Sensing voltage source %s for %s not found. BUG!" %
                                       (elem.source_id, elem.part_id))
                v = 0.
                v = v + x[elem.n1 - 1] if elem.n1 != 0 else v
                v = v - x[elem.n2 - 1] if elem.n2 != 0 else v
                tot_power = tot_power - v * elem.alpha * x[nv_1 + local_i_index, 0]
            elif isinstance(elem, devices.HVSource):
                try:
                    local_i_index = circ.find_vde_index(elem.source_id)
                except ValueError:
                    raise RuntimeError("Sensing voltage source %s for %s not found. BUG!" %
                                       (elem.source_id, elem.part_id))
                local_i_index2 = circ.find_vde_index(elem.part_id)
                tot_power = tot_power - elem.alpha*x[nv_1 + local_i_index, 0]* \
                                        x[nv_1 + local_i_index2, 0]
            elif circuit.is_elem_voltage_defined(elem):
                i_index = i_index + 1

        #op_info.append("TOTAL POWER: %e W\n" % (tot_power,))
        return op_keys, op_info, tot_power

    def write_to_file(self, filename=None):
        if filename is None and self.filename is None:
            # maybe warn the user here?
            return
        if filename is None:
            filename = self.filename
        if filename != 'stdout':
            fp = printing.open_utf8(filename+"info")
        else:
            fp = sys.stdout
        fp.write(self.timestamp+"\n")
        fp.write("ahkab v. "+__version__+" (c) 2006-2015 Giuseppe Venturini\n\n")
        fp.write("Operating Point (OP) analysis\n\n")
        fp.write("Netlist: %s\nTitle: %s\n" % (self.netlist_file, self.netlist_title))
        fp.write("At %.2f K\n" % (self.temp,))
        fp.write("Options:\n\tvea = %e\n\tver = %f\n\tiea = %e\n\tier = %f\n\tgmin = %e\n" \
                 % (self.vea, self.ver, self.iea, self.ier, self.gmin))
        fp.write("\nConvergence reached in %d iterations.\n" % (self.iterations,))
        fp.write("\n========\n")
        fp.write("RESULTS:\n")
        fp.write("========\n\n")
        vtable = self.get_table_array()
        fp.write(vtable+'\n')
        fp.write("\n========================\n")
        fp.write("ELEMENTS OP INFORMATION:\n")
        fp.write("========================\n\n")
        for k in sorted(self._op_info.keys()):
            t = printing.table(self._op_info[k], headers=self._op_keys[k][0])
            fp.write(t + '\n\n')
        fp.write('Total power dissipation: %g W\n\n' % self.tot_power)
        fp.flush()
        if filename != 'stdout':
            fp.close()
            # save to .op file
            self._add_data(self.x)

    def print_short(self):
        """Print a short, essential representation of the OP results"""
        table = []
        line = []
        for v in self.variables:
            line.append("%s: %g %s" % \
                        (v, self.results[v], self.units[v]))
            if len(line) == 5:
                table.append(line)
                line = []
        if len(line) > 0: # add the last line
            line += [""]*(5 - len(line))
            table.append(line)
        print(printing.table(table))

    @staticmethod
    def gmin_check(op2, op1):
        """Checks the differences between two sets of OP results.

        It is assumed that one set of results is calculated with Gmin, the other without.

        **Parameters:**

        op1, op2: op_solution instances
            the results vectors, interchangeable

        **Returns:**

        test_fail_variables : list
            The list of the variables that did not pass the test. They are extracted from
            the op_solution objects. If the check was passed, this is an empty list.
        """

        check_failed_vars = []
        for v in op1.variables:
            abserr = abs(op2.results[v] - op1.results[v])
            if op1.units[v] == 'V':
                if abserr > options.ver*max(abs(op1.results[v]),
                                            abs(op2.results[v])) + options.vea:
                    check_failed_vars.append(v)
            elif op1.units[v] == 'A':
                if abserr > options.ier*max(abs(op1.results[v]),
                                            abs(op2.results[v])) + options.iea:
                    check_failed_vars.append(v)
            else:
                print("Unrecognized unit... Bug.")
        return check_failed_vars

    def values(self):
        """Get all of the results set's variables values."""
        return np.squeeze(self.x).tolist()

    def items(self):
        vlist = []
        for j in range(self.x.shape[0]):
            vlist.append(self.x[j, 0])
        return list(zip(self.variables, vlist))

    # iterator methods
    def __iter__(self):
        self._iter_index = 0
        return self

    def next(self):
        if self._iter_index == len(self.variables):
            self._iter_index = 0
            raise StopIteration
        else:
            nxt = self.variables[self._iter_index], \
                   self.x[self._iter_index]
            self._iter_index += 1
        return nxt

    def __next__(self):
        return self.next()


class ac_solution(solution, _mutable_data):
    """AC results

    **Parameters:**

    circ : circuit instance
        the circuit instance of the simulated circuit
    start : float
       the AC sweep frequency start value, in Hz.
    stop : float
       the AC sweep frequency stop value, in Hz.
    points : int
       the AC sweep total points.
    stype : str
       the type of sweep, ``"LOG"``, ``"LIN"`` or arb. ``"POINTS"``.
    op : op_solution
       the linearization Operating Point used to compute the results.
    outfile: str
        the file to write the results to.  Use ``"stdout"`` to write to the
        standard output.
    """
    def __init__(self, circ, start, stop, points, stype, op, outfile):
        solution.__init__(self, circ, outfile)
        self.sol_type = "AC"
        self.linearization_op = op
        self.stype = stype
        self.ostart, self.ostop, self.opoints = start, stop, points

        self.variables += ["f"]
        self.units.update({"f": "Hz"})
        self.csv_headers = [self.variables[0]]

        nv_1 = circ.get_nodes_number() - 1 # numero di soluzioni di tensione (al netto del ref)
        for index in range(nv_1):
            varname = "V%s" % str(circ.nodes_dict[index + 1])
            self.variables += [varname]
            self.units.update({varname: "V"})
            if circ.is_int_node_internal_only(index+1):
                self.skip_nodes_list.append(index)

        for elem in circ:
            if circuit.is_elem_voltage_defined(elem):
                varname = "I(%s)" % elem.part_id.upper()
                self.variables += [varname]
                self.units.update({varname: "A"})

        for i in range(1, len(self.variables)):
            self.csv_headers.append("|%s|" % self.variables[i])
            self.csv_headers.append("arg(%s)" % self.variables[i])

    def _add_data(self, data):
        """Remember to call this method with REAL data - already split in ABS and PHASE."""
        csvlib.write_csv(self.filename, data, self.csv_headers, append=self._init_file_done)
        self._init_file_done = True

    def __str__(self):
        return ("<AC simulation results for '%s' (netlist %s). %s sweep, " +
                "from %g to %g rad/sec, %d points. Run on %s, data file " +
                "%s>") % (self.netlist_title, self.netlist_file, self.stype,
                          self.ostart, self.ostop, self.opoints, self.timestamp,
                          self.filename)

    def add_line(self, frequency, x):
        frequency = np.array([[frequency]])
        xsplit = np.zeros((x.shape[0]*2, 1))
        for i in range(x.shape[0]):
            xsplit[2*i, 0] = np.abs(x[i, 0])
            xsplit[2*i+1, 0] = np.angle(x[i, 0], deg=options.ac_phase_in_deg)

        data = np.concatenate((frequency, xsplit), axis=0)
        self._add_data(data)

    def get_x(self):
        return self[self.variables[0]]

    def get_xlabel(self):
        return self.variables[0]

    def asarray(self):
        """Return all data as a (possibly huge) python matrix."""
        ## data, headers, pos, EOF = csvlib.load_csv()
        data, headers, _, _ = csvlib.load_csv(self.filename, load_headers=[],
                                              nsamples=None, skip=0, verbose=0)
        cplx_data = None
        cplx_headers = []
        re1 = '\\|(.*?)\\|'
        rg = re.compile(re1, re.IGNORECASE|re.DOTALL)

        for i in range(len(headers)):
            if headers[i].upper() == self.variables[0].upper():
                if cplx_data is None:
                    cplx_data = np.array(data[i, :].reshape((1, -1)), dtype=np.complex_)
                else:
                    cplx_data = np.vstack((cplx_data, data[i, :].reshape((1, -1))))
            else:
                m = rg.search(headers[i])
                if m: # we got a |VAR|
                    var = m.group(1)
                    cplx_headers.append(var)
                    match_phase = ('arg(%s)' % var).upper()
                    ip = [h.upper() for h in headers].index(match_phase)
                    if cplx_data is None:
                        cplx_data = np.array(data[i, :]*
                                             np.exp(1j*data[ip,:]).reshape((1, -1)),
                                             dtype=np.complex_)
                    else:
                        cplx_data = np.vstack((cplx_data,
                                               (data[i, :]*
                                                np.exp(1j*data[ip, :])).reshape((1, -1))))
        return cplx_data

    # Access as a dictionary BY VARIABLE NAME:
    def __getitem__(self, name):
        """Get a specific variable, as from a dictionary."""
        if name.upper() != 'F':
            headers = ['|%s|' % name, 'arg(%s)' % name]
        else:
            headers = [name]
        try:
            # data, headers, pos, EOF = csvlib.load_csv()
            data, headers, _, _ = csvlib.load_csv(self.filename,
                                                  load_headers=headers,
                                                  nsamples=None, skip=0,
                                                  verbose=0)
        except ValueError:
            # raise the correct exception
            raise KeyError(name)
        if len(headers) == 2:
            data = data[0, :] * np.exp(1j*data[1, :])
        else:
            data = data.reshape((-1,))
        return data

    def get(self, name, default=None):
        """Get a solution by variable name."""
        try:
            data = self.__getitem__(name)
        except KeyError:
            return default
        return data

    def values(self):
        """Get all of the results set's variables values."""
        data = self.asarray()
        values = [np.real_if_close(data[0, :])]
        for i in range(1, data.shape[0]):
            values.append(data[i, :])
        return values

    def items(self):
        values = self.values()
        return zip(self.variables, values)

    # iterator methods
    def __iter__(self):
        self.iter_index = 0
        self.iter_data = self.values()
        self.iter_headers = self.variables
        return self

class dc_solution(solution, _mutable_data):
    """DC results

       **Parameters:**

       circ : circuit instance
           the simulated circuit.
       start : float
           the DC sweep start value.
       stop : float
           the DC sweep stop value.
       sweepvar : str
           the swept variable ``part_id``.
       stype : str
           the type of sweep, ``"LOG"``, ``"LIN"`` or arb. ``"POINTS"``.
       outfile : str
           the filename of the file where the results will be written.
           Use ``"stdout"`` to write to std output.
    """
    def __init__(self, circ, start, stop, sweepvar, stype, outfile):
        solution.__init__(self, circ, outfile)
        self.sol_type = "DC"
        self.start, self.stop = start, stop
        self.stype = stype

        nv_1 = circ.get_nodes_number() - 1 # numero di soluzioni di tensione (al netto del ref)
        self.variables = [sweepvar]
        self.units = case_insensitive_dict()
        if self.variables[0][0] == 'V':
            self.units.update({self.variables[0]:'V'})
        if self.variables[0][0] == 'I':
            self.units.update({self.variables[0]:'A'})

        for index in range(nv_1):
            varname = "V%s" % (str(circ.nodes_dict[index + 1]),)
            self.variables += [varname]
            self.units.update({varname:"V"})
            if circ.is_int_node_internal_only(index+1):
                self.skip_nodes_list.append(index)

        for elem in circ:
            if circuit.is_elem_voltage_defined(elem):
                varname = "I(%s)" % (elem.part_id.upper(),)
                self.variables += [varname]
                self.units.update({varname:"A"})

    def __str__(self):
        return ("<DC simulation results for '%s' (netlist %s). %s sweep of" +
                " %s from %g to %g %s. Run on %s, data file %s>") % \
               (self.netlist_title, self.netlist_file, self.stype,
                self.variables[0].upper(), self.start, self.stop,
                self.units[self.variables[0]], self.timestamp, self.filename)

    def add_op(self, sweepvalue, op):
        """A DC sweep is made of a set of OP points.

        This method adds an OP solution and
        its corresponding sweep value to the results set.
        """
        sweepvalue = np.array([[sweepvalue]])
        x = op.asarray()
        data = np.concatenate((sweepvalue, x), axis=0)
        self._add_data(data)

    def get_x(self):
        return self.get(self.variables[0])

    def get_xlabel(self):
        return self.variables[0]

class tran_solution(solution, _mutable_data):
    """Transient results

    **Parameters:**

    circ : circuit instance
        the circuit instance of the simulated circuit.
    tstart : float
        the transient simulation start time.
    tstop : float
        the transient simulation stop time.
    op : op_solution instance
        the Operating Point (OP) used to start the transient analysis.
    method : str
        the differentiation method employed.
    outfile : str
        the filename of the save file.
        Use "stdout" to write to the standard output.
    """
    def __init__(self, circ, tstart, tstop, op, method, outfile):
        solution.__init__(self, circ, outfile)
        self.sol_type = "TRAN"
        self.start_op = op
        self.tstart, self.tstop = tstart, tstop
        self.method = method

        self._lock = False

        nv_1 = circ.get_nodes_number() - 1 # numero di soluzioni di tensione (al netto del ref)
        self.variables = ["T"]
        self.units.update({"T":"s"})

        for index in range(nv_1):
            varname = ("V%s" % (str(circ.nodes_dict[index + 1]),)).upper()
            self.variables += [varname]
            self.units.update({varname:"V"})
            if circ.is_int_node_internal_only(index+1):
                self.skip_nodes_list.append(index)

        for elem in circ:
            if circuit.is_elem_voltage_defined(elem):
                varname = ("I(%s)" % (elem.part_id.upper(),)).upper()
                self.variables += [varname]
                self.units.update({varname:"A"})

    def __str__(self):
        return ("<TRAN simulation results for '%s' (netlist %s), from %g s to" +
                " %g s. Diff. method %s. Run on %s, data file %s>") % \
               (self.netlist_title, self.netlist_file, self.tstart, self.tstop,
                self.method, self.timestamp, self.filename)

    def add_line(self, time, x):
        """This method adds a solution and its corresponding time value to the results set.
        """
        if not self._lock:
            time = np.array([[time]])
            data = np.concatenate((time, x), axis=0)
            self._add_data(data)
        else:
            raise RuntimeError("Attempting to add values to a complete " +
                               "result set.")

    def lock(self):
        self._lock = True

    def get_x(self):
        return self.get(self.variables[0])

    def get_xlabel(self):
        return self.variables[0]


class pss_solution(solution, _mutable_data):
    """PSS results

    **Parameters:**

    circ : circuit instance
        the circuit instance of the simulated circuit.
    method : str
        the PSS algorithm employed.
    period : float
        the solution period.
    outfile : str
        the filename of the save file.
        Use "stdout" to write to the std output.

    .. note::

        Instantiating ``pss_solution`` creates an *empty* data set. Call
        :func:`set_results` to initialize its data.

    """
    def __init__(self, circ, method, period, outfile):
        solution.__init__(self, circ, outfile)
        self.sol_type = "PSS"
        self.period = period
        self.method = method

        # We have mixed current and voltage results
        nv_1 = circ.get_nodes_number() - 1 # numero di soluzioni di tensione (al netto del ref)
        self.variables = ["T"]
        self.units.update({"T":"s"})

        for index in range(nv_1):
            varname = "V%s" % (str(circ.nodes_dict[index + 1]),)
            self.variables += [varname]
            self.units.update({varname:"V"})
            if circ.is_int_node_internal_only(index+1):
                self.skip_nodes_list.append(index)

        for elem in circ:
            if circuit.is_elem_voltage_defined(elem):
                varname = "I(%s)" % (elem.part_id.upper(),)
                self.variables += [varname]
                self.units.update({varname:"A"})

    def __str__(self):
        return ("<PSS simulation results for '%s' (netlist %s), period %g s. " +
                "Method: %s. Run on %s, data file %s>") % \
               (self.netlist_title, self.netlist_file, self.period, self.method,
                self.timestamp, self.filename)

    def set_results(self, t, x):
        """Set the results in the data set

        .. note::

            * All the data are set at the same time for a PSS results set.
            * Instantiating ``pss_solution`` creates an empty data set.
            * This method should be called as soon as the data is available.

        **Parameters:**

        t : ndarray
            The time. The array should be 2D with shape ``(1, N)``.
        x : ndarray
            The data corresponding to the variables.
            The array should be 2D with shape ``(M, N)``, where ``M`` is the
            number of variables in the data set.

        """
        time = np.array(t)
        data = np.concatenate((time, x), axis=0)
        self._add_data(data)

    def asarray(self):
        allvalues, _, _, _ = csvlib.load_csv(self.filename, load_headers=[],
                                             nsamples=None, skip=0, verbose=0)
        return allvalues

    def get_x(self):
        return self.get(self.variables[0])

    def get_xlabel(self):
        return self.variables[0]

class symbolic_solution(object):
    """Symbolic results

    **Parameters:**

    results_dict : dict
        the results dict returned by ``sympy.solve()``,
    substitutions : dict
        the substitutions (dictionary) employed before solving,
    circ : circuit instance
        the circuit instance of the simulated circuit.
    outfile : str, optional
        the filename of the save file.
        Use ``"stdout"`` to write to the standard output.
    tf : bool, optional
        Transfer function flag: set this to ``True`` if this set of results
        corrsponds to a transfer function.  Defaults to ``False``.
    """
    def __init__(self, results_dict, substitutions, circ, outfile=None, tf=False):
        self.sol_type = "Symbolic"
        self.timestamp = time.strftime("%Y-%m-%d %H:%M:%S", time.gmtime())
        self.netlist_file = circ.filename
        self.netlist_title = circ.title
        self.substitutions = substitutions
        self.tf = tf

        # the keys are strings
        # self.symbols = map(str, results_dict.keys())
        self.results = case_insensitive_dict()
        for symbol, result in results_dict.items():
            self.results.update({str(symbol):result})

        self._symbols = list(results_dict.keys()) # keep them, they're useful
        for expr in list(results_dict.values()):
            if tf:
                expr = expr['gain']
            for symb in expr.atoms():
                if symb.is_Symbol and symb not in self._symbols:
                    self._symbols.append(symb)
        self.filename = outfile if outfile != 'stdout' else None
        if self.filename is not None:
            self.save()

    def as_symbol(self, variable):
        """Converts a string to the corresponding symbolic variable.

        This symbol may then be used by the user as an atom to construct
        new expressions, modify the results expressions or it can be passed
        to Sympy's functions.

        **Parameters:**

        variable : string
            The string that identifies the variable. Eg. ``'R1'`` for the variable
            corresponding to the resistance of the resistor ``R1``. Note that the
            case is disregarded and that the first letter defines the type of
            the element (resistor, capacitor...).

        **Returns:**

        symbol : Sympy symbol
            The corresponding symbol, if it exists in the result set.

        **Raises:**

        ValueError : exception
            In case no such symbol is found.
        """

        symbs = [x for x in self._symbols if x.name.lower() == variable.lower()]
        if len(symbs) == 0:
            raise ValueError("No symbol %s in the results set."%(variable,))
        else:
            return symbs[0]

    def as_symbols(self, spacedstr):
        """Convenience function to call :func:`as_symbol` multiple times.

        **Parameters:**

        spacedstr : string,
            A string containing several symbol identifiers separated by spaces.
            Eg. ``'R1 C2 L3'``.

        **Returns:**

        (s1, s2, ...) : tuple of Sympy symbol instances
            The symbols corresponding to the identifiers in the string supplied,
            ordered as the identifiers in the string.

        **Raises:**

        ValueError : exception
            In case any corresponding symbol is not found.
        """
        return list(map(self.as_symbol, spacedstr.split()))

    def save(self):
        """Write the results to disk.

        It is necessary first to set the ``filename`` attribute, indicating
        which file to write to.

        **Raises:**

        RuntimeError : exception
            If the `filename` attribute is not set.
        """
        if not self.filename:
            raise RuntimeError("Writing the results to file requires setting the \
                              'filename' attribute")
        with open(self.filename, 'wb') as fp:
            pickle.dump(self, fp, protocol=2)

    @staticmethod
    def load(filename):
        """Static method to load a symbolic solution from disk.

        **Parameters:**

        filename : str
            The filename corresponding to the file to load from.

        **Returns:**

        sol : symbolic solution instance
            The solution instance loaded from disk.

        .. warning::

            This method employs ``pickle.load``, which is to be used exclusively
            on trusted data. **Only load trusted simulation files!**

        """
        with open(filename, 'rb') as fp:
            asolution = pickle.load(fp)
        return asolution

    def __str__(self):
        str_repr = "Symbolic %s results for '%s' (netlist %s).\nRun on %s.\n" % \
                   ('simulation'*(not self.tf) + 'transfer function'*self.tf,
                    self.netlist_title, self.netlist_file, self.timestamp)
        keys = list(self.results.keys())
        keys.sort(key=str)
        if not self.tf:
            for key in keys:
                str_repr +=  str(key) + "\t = " + str(self.results[key]) + "\n"
        else:
            for key in keys:
                str_repr +=  str(key) + ":\n\t%s:\t%s\n" % \
                             ('gain', self.results[key]['gain'])
                if self.results[key]['gain'] == self.results[key]['gain0']:
                    continue
                str_repr +=  "\t%s:\t%s\n" % \
                             ('gain0', self.results[key]['gain0'])
                for sing in ('poles', 'zeros'):
                    if not len(self.results[key][sing]):
                        continue
                    str_repr += "\t%s:\n" % sing
                    for p in self.results[key][sing]:
                        str_repr +=  "\t\t" + str(p) + "\n"
        return str_repr

    # Access as a dictionary:
    def __len__(self):
        """Get the number of variables in the results set."""
        return len(self.results)

    def __getitem__(self, name):
        """Get a specific header, as from a dictionary."""
        return self.results[str(name).upper()]

    def get(self, name, default=None):
        """Get the solution corresponding to a variable."""
        name = str(name).upper()
        try:
            return self.results[name]
        except KeyError:
            return default

    def has_key(self, name):
        """Determine whether the result set contains a variable."""
        return str(name).upper() in self.results

    def __contains__(self, name):
        """Determine whether the result set contains a variable."""
        return str(name).upper() in self.results

    def keys(self):
        """Get all of the results set's variable's names."""
        return list(self.results.keys())

    def values(self):
        """Get all of the results set's variable's values."""
        return list(self.results.values())

    def items(self):
        """Get all solutions."""
        return list(self.results.items())

    # iterator methods
    def __iter__(self):
        self.iter_index = -1
        return self

    def __next__(self):
        """Iterator method."""
        if self.iter_index == len(list(self.results.keys())) - 1:
            self.iter_index = 0
            raise StopIteration
        else:
            self.iter_index += 1
        return list(self.results.keys())[self.iter_index], \
               self.results[self._symbols[self.iter_index]]

    def next(self):
        return self.__next__()

class pz_solution(solution, _mutable_data):
    """PZ results

    **Parameters:**

    circ : circuit instance
        the circuit instance of the simulated circuit.
    poles : sequence
        the circuit zeros
    zeros : sequence
        the circuit poles
    outfile : str
        the filename of the save file.
    """
    def __init__(self, circ, poles, zeros, outfile):
        solution.__init__(self, circ, outfile)
        self.sol_type = "PZ"
        self.poles = np.sort_complex(np.array(poles).reshape((-1,)))
        self.zeros = np.sort_complex(np.array(zeros).reshape((-1,)))
        data = np.vstack((self.poles.reshape((-1, 1)),
                          self.zeros.reshape((-1, 1))))
        if np.prod(self.poles.shape):
            for i in range(self.poles.shape[0]):
                self.variables += ['p%d' % i]
        if np.prod(self.zeros.shape):
            for i in range(self.zeros.shape[0]):
                self.variables += ['z%d' % i]
            for v in self.variables:
                self.units.update({v: "rad/s"})
            self.csv_headers = []
            for i in range(len(self.variables)):
                self.csv_headers.append("Re(%s)" % self.variables[i])
                self.csv_headers.append("Im(%s)" % self.variables[i])

            # save in Re/Im form
            sdata = data.reshape(-1).view(np.float_).reshape((-1, 1))
            self._add_data(sdata)

            # store local data too:
            self.data = case_insensitive_dict()
            for i in range(len(self.variables)):
                self.data.update({self.variables[i]: data[i, 0]})

    def _add_data(self, data):
        """Remember to call this method with REAL data - already split in RE and IM."""
        csvlib.write_csv(self.filename, data, self.csv_headers, append=self._init_file_done)
        self._init_file_done = True

    def __str__(self):
        return ("PZ simulation results for %s (netlist %s).\n" + \
               "Poles: %s\nZeros: %s") % \
               (self.netlist_title, self.netlist_file, list(self.poles),
                list(self.zeros))

    # Access as a dictionary BY VARIABLE NAME:
    def __getitem__(self, name):
        """Get a specific variable, as from a dictionary."""
        if name in self.data:
            return self.data[name]
        raise KeyError

    def get(self, name, default=None):
        try:
            return self.data[name]
        except KeyError:
            return default

    def has_key(self, name):
        """Determine whether the result set contains a variable."""
        return name in self.data

    def __contains__(self, name):
        """Determine whether the result set contains a variable."""
        return name in self.data

    def keys(self):
        """Get all of the results set's variable's names."""
        return self.data.keys()

    def values(self):
        """Get all of the results set's variable's values."""
        return self.data.values()

    def items(self):
        return self.data.items()

    # iterator methods
    def __iter__(self):
        # take into account that we increment first, then return
        # the value
        self.iter_index = -1
        return self

    def __next__(self):
        # redefine this or the superclass method will be used on
        # PY3, calling the superclass' next() and not *our* next
        # method below.
        return self.next()

    def next(self):
        if self.iter_index == len(self.variables)-1:
            self.iter_index = 0
            raise StopIteration
        else:
            self.iter_index += 1
        return self.variables[self.iter_index], \
               self.data[self.variables[self.iter_index]]

class case_insensitive_dict(object):
    """A dictionary that uses case-insensitive strings as keys.
    """
    def __init__(self):
        self._dict = {}

    def __len__(self):
        """Get the number of elements in the set."""
        return len(self._dict)

    def __str__(self):
        rpr = "{"
        keys = list(self._dict.keys())
        for i in range(len(keys)):
            rpr += "%s: %s" % (keys[i], self._dict[keys[i]])
            if i < len(keys) - 1:
                rpr += ", "
            else:
                rpr += "}"
        return rpr

    def __getitem__(self, name):
        """Get a specific variable, as from a dictionary."""
        keys = list(self._dict.keys())
        try:
            i = [k.upper() for k in keys].index(text_type(name).upper())
        except ValueError:
            raise KeyError(name)
        return self._dict[keys[i]]

    def get(self, name, default=None):
        """Given the case-insensitive string key ``name``, return its corresponding value.

        If not found, return ``default``.
        """
        try:
            keys = list(self._dict.keys())
            i = [k.upper() for k in keys].index(name.upper())
        except ValueError:
            return default
        return self._dict[keys[i]]

    def has_key(self, name):
        """Determine whether the result set contains the variable ``name``."""
        return name.upper() in [k.upper() for k in list(self._dict.keys())]

    def __contains__(self, name):
        """Determine whether the result set contains a variable."""
        return name.upper() in [k.upper() for k in list(self._dict.keys())]

    def keys(self):
        """Get all keys"""
        return list(self._dict.keys())

    def values(self):
        """Get all values"""
        return list(self._dict.values())

    def items(self):
        """Get all keys and values pairs"""
        return list(self._dict.items())

    def update(self, adict):
        """Update the dictionary contents with the mapping in the dictionary ``adict``.
        """
        return self._dict.update(adict)

    # iterator methods
    def __iter__(self):
        """Iterator"""
        return self._dict.__iter__()



# -*- coding: utf-8 -*-
# csvlib.py
# Implementation of routines for CSV data I/O
# Copyright 2012 Giuseppe Venturini
#
# This file is part of the ahkab simulator.
#
# Ahkab is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# Ahkab is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License v2
# along with ahkab.  If not, see <http://www.gnu.org/licenses/>.
"""The ``csvlib`` module contains common routines for handling
Comma Separated Values (CSV) or Tab Separated Values (TSV) files.

Functions:

1. CSV write/load:

  * :func:`write_csv`
  * :func:`load_csv`

2. MISC utilities

  * :func:`get_headers`
  * :func:`write_headers`
  * :func:`get_headers_index`

The separator can be selected setting:

::

   csvlib.SEPARATOR = '\\t' # default value


"""

# Additionally the following internal functions are
# available:
#3. Internal routines
#    _get_fp(filename, mode='rb')
#    _close_fp(fp, filename)

from __future__ import (unicode_literals, absolute_import,
                        division, print_function)

import io
import sys
import copy
import os
import numpy as np

from . import options

SEPARATOR = u"\t"


def write_csv(filename, data, headers, append=False):
    """Writes data in CVS format to filename.

    The headers have to be ordered according to the data order.

    **Parameters:**

    filename : string
        the path to the file to be written.
        Use 'stdout' to write to stdout

    data : ndarray
        The data to be written. Notice that variables are swept across *rows*,
        time samples are swept along *columns*.
        Or equivalently: ``data[variable_index, sample_number]``

    headers : list of strings
        the signal names, ordered so that ``headers[i]`` corresponds to 
        ``data[i, :]``.

    append : bool, optional
        If False, the file (if it exists) will be rewritten, otherwise
        it will be appended to.

    """

    mode = 'ab' if append else 'wb'
    fp = _get_fp(filename, mode)

    if not data.shape[0] == len(headers):
        print("(W): write_csv(): data and headers don't match. Continuing anyway.")
        print("DATA: " + str(data.shape) + " headers length: " + str(len(headers)))

    headers = SEPARATOR.join(headers) if not append else ""
    np.savetxt(fp, data.T, delimiter=SEPARATOR, header=headers, comments='#')

    _close_fp(fp, filename)


def write_headers(filename, headers):
    """Writes headers in CVS format to filename.

    **Parameters:**

    filename : string
        the path to the file to be written.
        Use 'stdout' to write to stdout.

    headers : a list of strings
        the signal names, ordered.

    """
    fp = _get_fp(filename, mode="wb")
    headers = copy.copy(headers)
    if not headers[0][0] == '#':
        headers[0] = '#' + headers[0]
    for hi in range(len(headers)):
        fp.write(headers[hi])
        if hi < len(headers) - 1:
            fp.write(SEPARATOR)
        else:
            fp.write("\n")


def _get_fp(filename, mode="r"):
    if filename == 'stdout' or filename == '-' or filename == sys.stdout:
        if mode == 'w' or mode == 'a' or mode == 'wb' or mode == 'ab':
            fp = sys.stdout if not hasattr(sys.stdout, 'buffer') else sys.stdout.buffer
        else:
            print("(EE) Mode %s is not supported for stdout." % (mode,))
            fp = None
    else:
        fp = io.open(filename, mode)#, encoding=options.encoding)
    return fp


def _close_fp(fp, filename):
    try:
        fp.flush()
    except IOError:
        pass
    if filename == 'stdout':
        pass
    else:
        fp.close()


def get_headers_index(headers, load_headers=None, verbose=3):
    """Creates a list of integers. Each element in the list is the COLUMN index
    of the signal according to the supplied headers.

    **Parameters:**

    headers : list of strings,
        the signal names, as returned by :func:`get_headers`.

    load_headers : list, optional
        The headers for the data to be loaded. If not provided, all indeces will
        be returned.

    **Returns:**

    the header indeces : a list of int.

    """
    if load_headers is None or not len(load_headers):
        return list(range(len(headers)))
    his = []
    lowcase_headers = [i.lower() for i in headers]

    for lh in load_headers:
        try:
            his = his + [lowcase_headers.index(lh.lower())]
        except ValueError:
            if verbose:
                print("(W): header " + lh + " not found. Skipping.")
    return his


def get_headers(filename):
    """Reads the signals inside a file.

    The order of the signals in the list corresponds to the order of the
    signals in the file.

    *Parameters:*

    filename : string
        the path to the file from which the header is to be read

    **Returns:**

    headers : list of strings.

    """

    fp = _get_fp(filename, 'r')
    headers = None
    line = ""
    while line == "":
        line = fp.readline()
        line = line.strip()
        if line[0] == '#':
            line = line[1:]
    headers = line.split(SEPARATOR)
    return headers


def load_csv(filename, load_headers=None, nsamples=None, skip=0, verbose=3):
    """Reads data in CVS format from filename.

    Supports:

    * selective signal loading,
    * loading up to a certain number of samples,
    * skipping to a certain line, to allow incremental reading of big files.

    **Parameters:**

    filename : string
        the path to the file to be read.

    load_headers : list of strings, optional
        Each one being a signal to be loaded. An empty list (or None) is 
        interpreted as "read all signals".

    nsamples : int, optional
        The number of samples to be read for each signal. If ``None``,
        read all available samples.

    skip : int, optional
        The index of the first sample to be read. Default: 0

    **Returns:**

    data : ndarray 
        The data, ordered according to the order of ``load_headers``
        (or the order on file if ``load_headers`` was empty),

    headers : list of strings
        the names of the signals read from file,

    pos : int
        position of the last sample read +1, referred to the
        sample #0 in the file.

    EOF : bool
        A flag set to true is all the data in the file were read.

    """

    if filename == 'stdout':
        print("Can't load data from stdout.")
        return None, None, None, None

    headers = get_headers(filename)
    his = get_headers_index(headers, load_headers, verbose=verbose)
    if load_headers and len(his) != len(load_headers):
        raise ValueError("Specified header not found")

    fp = _get_fp(filename)
    data = np.loadtxt(fp, delimiter=SEPARATOR, usecols=his, unpack=True, skiprows=skip, ndmin=2)
    _close_fp(fp, filename)

    # prepare return values
    EOF = (nsamples is None) or (nsamples == data.shape[1])
    if nsamples is not None:
        data = data[:, :min(nsamples, data.shape[1])]
    pos = skip + data.shape[1]
    headers = list(map(headers.__getitem__, his))

    return data, headers, pos, EOF



# -*- coding: iso-8859-1 -*-
# bfpss.py
# Brute-force PSS analysis module
# Copyright 2006 Giuseppe Venturini

# This file is part of the ahkab simulator.
#
# Ahkab is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# Ahkab is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License v2
# along with ahkab.  If not, see <http://www.gnu.org/licenses/>.

"""Brute-force periodic steady state analysis module"""

from __future__ import (unicode_literals, absolute_import,
                        division, print_function)

import sys
import numpy as np
import numpy.linalg
import scipy
import scipy.sparse

from . import circuit
from . import dc_analysis
from . import devices
from . import implicit_euler
from . import options
from . import printing
from . import results
from . import ticker
from . import transient
from . import utilities


def bfpss(circ, period, step=None, points=None, autonomous=False, x0=None,
          mna=None, Tf=None, D=None, outfile='stdout',
          vector_norm=lambda v: max(abs(v)), verbose=3):
    """Performs a PSS analysis employing the 'brute-force' algorithm

    The time step is constant and IE will be used as DF.

    **Parameters:**

    circ : Circuit instance
        the circuit to be simulated

    period : float
        the period of the solution

    step : float, optional
        the time step between consecutive points, it will be calculated
        if not provided

    points : int, optional
        the number of points to be used to sample one period, it will be
        calculated if not provided

    autonomous : bool, optional
        Is the circuit clocked or autonomously oscillating?
        With the current implementation, setting ``autonomous=True``
        will result in an exception being raised, autonomous circuits are
        not supported

    x0 : ndarray, optional
        The initial guess to be used. (Experimental, needs work.)

    mna, D, Tf : ndarrays, optional
        The matrices describing the circuit may be supplied to speed up
        the solution, if available. If not supplied, they will be
        automatically calculated.

    vector_norm : function, optional
       The norm to be employed in the convergence checks.
       Defaults to the Inf norm.

    outfile : str, optional
        the output filename. Defaults to ``'stdout'``.

    verbose : int, optional
        Verbosity level on a scale from 0 (silent) to 6 (very verbose).
        The ``verbose`` flag is automatically set is to zero
        if ``datafilename == 'stdout'``

    .. note::

        ``step`` and ``points`` are mutually exclusive options:

        * if ``step`` is specified, the number of points will be
          automatically determined.
        * if ``points`` is set, the step will be automatically determined.
        * if none of them is set, ``options.bfpss_default_points`` will
          be used as value for ``points`` and ``step`` computed accordingly.

    **Returns:**

    sol : results.pss_solution
        The simulation results

    """
    if outfile == "stdout":
        verbose = 0

    printing.print_info_line(
        ("Starting periodic steady state analysis:", 3), verbose)
    printing.print_info_line(("Method: brute-force", 3), verbose)

    if mna is None or Tf is None:
        (mna, Tf) = dc_analysis.generate_mna_and_N(circ, verbose=verbose)
        mna = utilities.remove_row_and_col(mna)
        Tf = utilities.remove_row(Tf, rrow=0)
    elif not mna.shape[0] == Tf.shape[0]:
        printing.print_general_error(
            "mna matrix and N vector have different number of rows.")
        sys.exit(0)

    if D is None:
        D = transient.generate_D(circ, [mna.shape[0], mna.shape[0]])
        D = utilities.remove_row_and_col(D)
    elif not mna.shape == D.shape:
        printing.print_general_error(
            "mna matrix and D matrix have different sizes.")
        raise ValueError

    (points, step) = utilities.check_step_and_points(step, points, period,
                                           options.bfpss_default_points)

    n_of_var = mna.shape[0]
    locked_nodes = circ.get_locked_nodes()
    tick = ticker.ticker(increments_for_step=1)
    sparse = n_of_var*points > options.dense_matrix_limit

    CMAT = _build_CMAT(mna, D, step, points, tick, n_of_var=n_of_var,
                      sparse=sparse, verbose=verbose)

    x = _build_x(mna, step, points, tick, x0=x0, n_of_var=n_of_var,
                 verbose=verbose)

    Tf = _build_Tf(Tf, points, tick, n_of_var=n_of_var, verbose=verbose)

    # time variable component: Tt this is always the same in each iter.
    # So we build it once for all
    # It holds all time-dependent sources (both V/I).
    Tt = _build_Tt(circ, points, step, tick, n_of_var=n_of_var,
                   verbose=verbose)

    # Indices to differentiate between currents and voltages in the
    # convergence check
    nv_indices = []
    ni_indices = []
    nv_1 = circ.get_nodes_number() - 1
    ni = n_of_var - nv_1
    for i in range(points):
        nv_indices += (i * mna.shape[0] * np.ones(nv_1) + \
                      np.arange(nv_1)).tolist()
        ni_indices += (i * mna.shape[0] * np.ones(ni) + \
                      np.arange(nv_1, n_of_var)).tolist()

    converged = False

    printing.print_info_line(("Solving... ", 3), verbose, print_nl=False)
    tick.reset()
    tick.display(verbose > 2)
    if sparse:
        J = scipy.sparse.lil_matrix(CMAT.shape)
    else:
        J = np.zeros(CMAT.shape)
    T = np.zeros((CMAT.shape[0], 1))
    # td is a np array that will hold the damping factors
    td = np.zeros((points, 1))
    iteration = 0  # newton iteration counter

    while True:
        if iteration:  # the first time are already all zeros
            if sparse:
                J = scipy.sparse.lil_matrix(CMAT.shape)
            else:
                J[:, :] = 0
            T[:, 0] = 0
            td[:, 0] = 0
        for index in range(1, points):
            for elem in circ:
                # build all dT(xn)/dxn (stored in J) and T(x)
                if elem.is_nonlinear:
                    oports = elem.get_output_ports()
                    for opindex in range(len(oports)):
                        dports = elem.get_drive_ports(opindex)
                        v_ports = []
                        for dpindex in range(len(dports)):
                            dn1, dn2 = dports[dpindex]
                            # build v: remember we trashed the
                            # 0 row and 0 col of mna -> -1
                            v = 0
                            if dn1:
                                v = v + x[index * n_of_var + dn1 - 1, 0]
                            if dn2:
                                v = v - x[index * n_of_var + dn2 - 1, 0]
                            v_ports.append(v)
                        # all drive ports are ready.
                        n1, n2 = oports[opindex][0], oports[opindex][1]
                        if n1:
                            T[index * n_of_var + n1 - 1, 0] = T[index * n_of_var + n1 - 1, 0] + \
                                                              elem.i(opindex, v_ports)
                        if n2:
                            T[index * n_of_var + n2 - 1, 0] = T[index * n_of_var + n2 - 1, 0] - \
                                                              elem.i(opindex, v_ports)
                        for dpindex in range(len(dports)):
                            dn1, dn2 = dports[dpindex]
                            if n1:
                                if dn1:
                                    J[index * n_of_var + n1 - 1, index * n_of_var + dn1 - 1] = \
                                        J[index * n_of_var + n1 - 1, index * n_of_var + dn1 - 1] + \
                                            elem.g(opindex, v_ports, dpindex)
                                if dn2:
                                    J[index * n_of_var + n1 - 1, index * n_of_var + dn2 - 1] =\
                                        J[index * n_of_var + n1 - 1, index * n_of_var + dn2 - 1] - 1.0 * \
                                            elem.g(opindex, v_ports, dpindex)
                            if n2:
                                if dn1:
                                    J[index * n_of_var + n2 - 1, index * n_of_var + dn1 - 1] = \
                                        J[index * n_of_var + n2 - 1, index * n_of_var + dn1 - 1] - 1.0 * \
                                            elem.g(opindex, v_ports, dpindex)
                                if dn2:
                                    J[index * n_of_var + n2 - 1, index * n_of_var + dn2 - 1] =\
                                        J[index * n_of_var + n2 - 1, index * n_of_var + dn2 - 1] + \
                                            elem.g(opindex, v_ports, dpindex)

        J = J + CMAT
        residuo = CMAT*x + T + Tf + Tt
        if sparse:
            J = scipy.sparse.csc_matrix(J)
            lu = scipy.sparse.linalg.splu(J)
            dx = lu.solve(-residuo)
        else:
            dx = np.linalg.solve(J, -residuo)
        # td
        for index in range(points):
            td[index, 0] = dc_analysis.get_td(dx[index*n_of_var:(index + 1)*n_of_var,
                                                 0].reshape(-1, 1),
                                              locked_nodes, n=-1)
        x = x + min(abs(td))[0] * dx
        # convergence check
        converged = _convergence_check(dx, x, nv_indices, ni_indices,
                                      vector_norm)
        if converged:
            break
        tick.step()

        if options.bfpss_max_nr_iter and iteration == options.bfpss_max_nr_iter:
            printing.print_general_error("Hit BFPSS_MAX_NR_ITER (" +
                                         str(options.bfpss_max_nr_iter) +
                                         "), iteration halted.")
            converged = False
            break
        else:
            iteration = iteration + 1

    tick.hide(verbose > 2)
    if converged:
        printing.print_info_line(("done.", 3), verbose)
        t = np.arange(points) * step
        t = t.reshape((1, points))
        x = x.reshape((points, n_of_var))
        sol = results.pss_solution(circ=circ, method="brute-force",
                                   period=period, outfile=outfile)
        sol.set_results(t, x.T)
    else:
        print("failed.")
        sol = None
    return sol


def _convergence_check(dx, x, nv_indices, ni_indices, vector_norm):
    """Perform a convergence check using the specified vector norm"""
    # sometimes something diverges... look out
    if vector_norm(dx) is np.nan:
        raise OverflowError
    dxc = np.array(dx)
    xc = np.array(x)
    ret = (vector_norm(dxc[nv_indices]) < options.ver * \
           vector_norm(xc[nv_indices]) + options.vea) and \
          (not len(ni_indices) or \
           vector_norm(dxc[ni_indices]) < options.ier * \
           vector_norm(xc[ni_indices]) + options.iea)
    return ret


def _build_CMAT(mna, D, step, points, tick, n_of_var=None, sparse=False, verbose=3):
    if n_of_var is None:
        n_of_var = mna.shape[0]
    printing.print_info_line(("Building CMAT (%dx%d)... " %
                             (n_of_var*points, n_of_var*points), 5), verbose, print_nl=False)
    tick.reset()
    tick.display(verbose > 2)
    C1, C0 = implicit_euler.get_df_coeff(step)
    I = np.eye(n_of_var)
    M = mna + C1*D
    N = C0*D
    if sparse:
        CMAT = scipy.sparse.lil_matrix((n_of_var*points, n_of_var*points))
    else:
        CMAT = np.zeros((n_of_var*points, n_of_var*points))
    for li in range(points):  # li = line index
        for ci in range(points):
            if li == 0:
                if ci == 0:
                    temp = I
                elif ci == points - 1:
                    temp = -I
                else:
                    continue  # temp = Z
            else:
                if ci == li:
                    temp = M
                elif ci == li - 1:
                    temp = N
                else:
                    continue  # temp = Z
            CMAT = utilities.set_submatrix(row=li*n_of_var, col=ci*n_of_var,
                                           dest_matrix=CMAT, source_matrix=temp)
        tick.step()
    tick.hide(verbose > 2)
    if sparse:
        CMAT = scipy.sparse.coo_matrix(CMAT)
    printing.print_info_line(("done.", 5), verbose)

    return CMAT


def _build_x(mna, step, points, tick, x0=None, n_of_var=None, verbose=3):
    if n_of_var is None:
        n_of_var = mna.shape[0]
    printing.print_info_line(("Building x...", 5), verbose, print_nl=False)
    tick.reset()
    tick.display(verbose > 2)
    x = np.zeros((points * n_of_var, 1))
    if x0 is not None:
        if isinstance(x0, results.op_solution):
            x0 = x0.asarray()
        if x0.shape[0] != n_of_var:
            print("Warning x0 has the wrong dimensions. Using all 0s.")
        else:
            for index in range(points):
                x = utilities.set_submatrix(row=index*n_of_var, col=0,
                                            dest_matrix=x, source_matrix=x0)
                tick.step()

    tick.hide(verbose > 2)
    printing.print_info_line(("done.", 5), verbose)
    return x


def _build_Tf(sTf, points, tick, n_of_var, verbose=3):
    printing.print_info_line(("Building Tf...", 5), verbose, print_nl=False)
    tick.reset()
    tick.display(verbose > 2)
    Tf = np.zeros((points * n_of_var, 1))

    for index in range(1, points):
        Tf = utilities.set_submatrix(row=index*n_of_var, col=0, dest_matrix=Tf,
                                     source_matrix=sTf)
        tick.step()
    tick.hide(verbose > 2)
    printing.print_info_line(("done.", 5), verbose)
    return Tf


def _build_Tt(circ, points, step, tick, n_of_var, verbose=3):
    nv = circ.get_nodes_number()
    printing.print_info_line(("Building Tt...", 5), verbose, print_nl=False)
    tick.reset()
    tick.display(verbose > 2)
    Tt = np.zeros((points * n_of_var, 1))
    for index in range(1, points):
        v_eq = 0
        time = index * step
        for elem in circ:
            if (isinstance(elem, devices.VSource) or isinstance(elem, devices.ISource)) and elem.is_timedependent:
                if isinstance(elem, devices.VSource):
                    Tt[index * n_of_var + nv - 1 + v_eq, 0] = - \
                        1.0 * elem.V(time)
                elif isinstance(elem, devices.ISource):
                    if elem.n1:
                        Tt[index * n_of_var + elem.n1 - 1, 0] = \
                            Tt[index * n_of_var + elem.n1 - 1, 0] + \
                            elem.I(time)
                    if elem.n2:
                        Tt[index * n_of_var + elem.n2 - 1, 0] = \
                            Tt[index * n_of_var + elem.n2 - 1, 0] - \
                                elem.I(time)
            if circuit.is_elem_voltage_defined(elem):
                v_eq = v_eq + 1
            # print Tt[index*n_of_var:(index+1)*n_of_var]
        tick.step()
    tick.hide(verbose > 2)
    printing.print_info_line(("done.", 5), verbose)

    return Tt



# -*- coding: iso-8859-1 -*-
# dc_guess.py
# DC guess functions
# Copyright 2007 Giuseppe Venturini

# This file is part of the ahkab simulator.
#
# Ahkab is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# Ahkab is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License v2
# along with ahkab.  If not, see <http://www.gnu.org/licenses/>.

"""This module provides the :func:`get_dc_guess` method, used to
compute a starting point to initialize a Newton-Rhapson solver.

Module reference
################

"""

from __future__ import (unicode_literals, absolute_import,
                        division, print_function)

import sys

import numpy as np
import numpy.linalg

from . import circuit
from . import utilities


def get_dc_guess(circ, verbose=3):
    """Build a DC guess from circuit inspection.

    Notice that OP analysis will call this method on the users' behalf if not
    instructed not to do so.

    A element can suggest its guess through the ``elem.dc_guess`` field.
    If the field is not set, or not available, no information on the most
    likely biasing voltage is assumed.

    **Parameters:**

    circ : Circuit instance
        The circuit instance the guess is being computed for.

    verbose : int, optional
        The verbosity level (from 0 silent to 6 debug). Defaults to 3, medium
        verbosity.

    **Returns:**

    dcg : ndarray or None
        The DC guess, in numpy array form, or ``None``, if it was not possible
        to compute a meaningful guess.
    """
    if verbose:
        sys.stdout.write("Calculating guess: ")
        sys.stdout.flush()

    # A DC guess has meaning only if the circuit has NL elements
    if not circ.is_nonlinear():
        if verbose:
            print("skipped. (linear circuit)")
        return None

    if verbose > 3:
        print("")

    nv = circ.get_nodes_number()
    M = np.zeros((1, nv))
    T = np.zeros((1, 1))
    index = 0
    v_eq = 0  # number of current equations
    one_element_with_dc_guess_found = False

    for elem in circ:
        # In the meanwhile, check how many current equations are
        # required to solve the circuit
        if circuit.is_elem_voltage_defined(elem):
            v_eq = v_eq + 1
        # This is the main focus: build a system of equations (M*x = T)
        if hasattr(elem, "dc_guess") and elem.dc_guess is not None:
            if not one_element_with_dc_guess_found:
                one_element_with_dc_guess_found = True
            if elem.is_nonlinear:
                port_index = 0
                for (n1, n2) in elem.ports:
                    if n1 == n2:
                        continue
                    if index:
                        M = utilities.expand_matrix(M, add_a_row=True,
                                                    add_a_col=False)
                        T = utilities.expand_matrix(T, add_a_row=True,
                                                    add_a_col=False)
                    M[index, n1] = +1
                    M[index, n2] = -1
                    T[index] = elem.dc_guess[port_index]
                    port_index = port_index + 1
                    index = index + 1
            else:
                if elem.n1 == elem.n2:
                    continue
                if index:
                    M = utilities.expand_matrix(M, add_a_row=True,
                                                add_a_col=False)
                    T = utilities.expand_matrix(T, add_a_row=True,
                                                add_a_col=False)
                M[index, elem.n1] = +1
                M[index, elem.n2] = -1
                T[index] = elem.dc_guess[0]
                index = index + 1

    if verbose == 5:
        print("DBG: get_dc_guess(): M and T, no reduction")
        print(M)
        print(T)
    M = utilities.remove_row_and_col(M, rrow=10 * M.shape[0], rcol=0)

    if not one_element_with_dc_guess_found:
        if verbose == 5:
            print("DBG: get_dc_guess(): no element has a dc_guess")
        elif verbose <= 3:
            print("skipped.")
        return None

    # We wish to find the linearly dependent lines of the M matrix.
    # The matrix is made by +1, -1, 0 elements.
    # Hence, if two lines are linearly dependent, one of these equations
    # has to be satisfied: (L1, L2 are two lines)
    # L1 + L2 = 0 (vector)
    # L2 - L1 = 0 (vector)
    # This is tricky, because I wish to remove lines of the matrix while
    # browsing it.
    # We browse the matrix by line from bottom up and compare each line
    # with the upper lines. If a linearly dep. line is found, we remove
    # the current line.
    # Then break from the loop, get the next line (bottom up), which is
    # the same we were considering before; compare with the upper lines..
    # Not optimal, but it works.
    for i in range(M.shape[0] - 1, -1, -1):
        for j in range(i - 1, -1, -1):
            # print i, j, M[i, :], M[j, :]
            dummy1 = M[i, :] - M[j, :]
            dummy2 = M[i, :] + M[j, :]
            if not dummy1.any() or not dummy2.any():
                # print "REM:", M[i, :]
                M = utilities.remove_row(M, rrow=i)
                T = utilities.remove_row(T, rrow=i)
                break
    if verbose == 5:
        print("DBG: get_dc_guess(): M and T, after removing LD lines")
        print(M)
        print(T)

    # Remove empty columns:
    # If a column is empty, we have no guess regarding the corresponding
    # node. It makes the matrix singular. -> Remove the col & remember
    # that we are _not_ calculating a guess for it.
    removed_index = []
    for i in range(M.shape[1] - 1, -1, -1):
        if not M[:, i].any():
            M = utilities.remove_row_and_col(M, rrow=M.shape[0], rcol=i)
            removed_index.append(i)

    if verbose > 3:
        print("DBG: get_dc_guess(): M and T, after removing empty columns.")
        print(M)
        print("T\n", T)

    # Now, we have a set of equations to be solved.
    # There are three cases:
    # 1. The M matrix has a different number of rows and columns.
    #    We use the Moore-Penrose matrix inverse to get
    #    the shortest length least squares solution to the problem
    #          M*x + T = 0
    # 2. The matrix is square.
    #    It seems that if the circuit is not pathological,
    #    we are likely to find a solution (the matrix has det != 0).
    #    I'm not sure about this though.

    if M.shape[0] != M.shape[1]:
        Rp = np.dot(np.linalg.pinv(M), T)
    else:  # case M.shape[0] == M.shape[1], use normal
        if np.linalg.det(M) != 0:
            try:
                Rp = np.dot(np.linalg.inv(M), T)
            except np.linalg.linalg.LinAlgError:
                eig = np.linalg.eig(M)[0]
                cond = abs(eig).max() / abs(eig).min()
                if verbose:
                    print("cond=" + str(cond) + ". No guess.")
                return None
        else:
            if verbose:
                print("Guess matrix is singular. No guess.")
            return None

    # Now we want to:
    # 1. Add voltages for the nodes for which we have no clue to guess.
    # 2. Append to each vector of guesses the values for currents in
    #    voltage defined elem.
    # Both them are set to 0
    for index in removed_index:
        Rp = np.concatenate((np.concatenate((Rp[:index, 0].reshape((-1, 1)),
                                             np.zeros((1, 1))), axis=0),
                             Rp[index:, 0].reshape((-1, 1))), axis=0)
    # add the 0s for the currents due to the voltage defined
    # elements (we have no guess for those...)
    if v_eq > 0:
        Rp = np.concatenate((Rp, np.zeros((v_eq, 1))), axis=0)

    if verbose and verbose < 4:
        print("done.")
    if verbose > 3:
        print("Guess:")
        print(Rp)

    return Rp



# -*- coding: iso-8859-1 -*-
# trap.py
# Trap DF (with a second order prediction)
# Copyright 2006 Giuseppe Venturini

# This file is part of the ahkab simulator.
#
# Ahkab is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# Ahkab is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License v2
# along with ahkab.  If not, see <http://www.gnu.org/licenses/>.

"""
This file implements the Trapezoidal (TRAP) Differentiation Formula (DF) and a
second order prediction formula.

Module reference
----------------
"""

from __future__ import (unicode_literals, absolute_import,
                        division, print_function)

import numpy as np
from numpy.linalg import inv

from .py3compat import range_type

order = 2


def is_implicit():
    """Is this Differentiation Formula (DF) implicit?

    **Returns:**

    isit : boolean
        In this case, that's ``True``.
    """
    return True


def get_required_values():
    """Get info regarding what values are needed by the DF

    **Returns:**

    tpl : tuple of tuples
        A tuple of two tuples.

        * The first tuple indicates what past values of the unknown are needed
          for the DF.
        * The second tuple indicates what past values of the unknown are needed
          for the prediction method.

        In particular, each of the sub-tuples is built this way:
        ::

            (max_order_of_x, max_order_of_dx)

        Where both the values are either ``int``, or ``None``.
        If ``max_order_of_x`` is set to an integer value :math:`k`, the DF needs
        all the :math:`x_{n-i}` values of x, for all :math:`0 \\le i \\le k`. In
        the previous text, :math:`x_{n-i}` is the value the :math:`x` array
        assumed :math:`i` steps before the one we are considering for the
        derivative.

        Similar considerations apply to ``max_order_of_dx``, but regard rather
        :math:`dx_n/dt` instead of :math:`x_n`.

        If any of the values is set to ``None``, it is to be assume that no
        value is required.

    The first array has to be used if no prediction is required, the second are the values needed for prediction.
    """
    # we need x(n) and dx(n)/dt
    return ((0, 0), (2, None))


def has_ff():
    """Has the method a Forward Formula for prediction?

    **Returns:**

    doesit : bool
        In this particular case, this function always returns ``True``.
    """
    return True


def get_df(pv_array, suggested_step, predict=True):
    """Get the coefficients for DF and prediction formula

    **Parameters:**

    pv_array : sequence of sequences
        Each element of ``pv_array`` must be of the form:

        ::

            (time, x, derivate(x))

        In particular, the :math:`k` element of `pv_array` contains the values
        of:

        * :math:`t_{n-k}` (the time),
        * :math:`x_{n-k}`,
        * :math:`dx_{n-k}/dt`

        evaluated :math:`k` time steps before the current one, labeled
        :math:`n+1`.

        How many samples are necessary is given by
        :func:`ahkab.trap.get_required_values`.

        Values that are not needed may be set to ``None``, as they will be
        disregarded.
    suggested_step : float
        The step that will be used for the current iteration, provided the error
        will be deemed acceptable.
    predict : bool, optional
        Whether a prediction for :math:`x_n` is needed as well or not. Defaults
        to ``True``.

    **Returns:**

    ret : tuple
        The return value has the form:

        ::

            (C1, C0, x_lte_coeff, predict_x, predict_lte_coeff)

        The derivative may be written as:

        .. math::

            d(x(n+1))/dt = C1 x(n+1) + C0

        `x_lte_coeff` is the coefficient of the Local Truncation Error,
        `predict_x` is the predicted value for :math:`x` and `predict_lte_coeff`
        is the LTE coefficient for the prediction.

    :raises ValueError: if the `pv_array` is malformed.
    """

    # our method needs x(n) dx(n)/dt
    if len(pv_array[0]) != 3:
        raise ValueError('trap.get_df got a pv_array of wrong dimensions')
    if pv_array[0][1] is None or pv_array[0][2] is None:
        raise ValueError('trap.get_df got a pv_array that does not contain'
                         ' required values')
    if predict and (len(pv_array) < 2 or pv_array[0][1] is None or
        pv_array[1][1] is None or pv_array[2][1] is None):
        raise ValueError('trap.get_df got predict=True but pv_array that does'
                         ' not contain the required values')

    C1 = 2.0 / suggested_step
    C0 = -1.0 * (2.0 / suggested_step * pv_array[0][1] + pv_array[0][2])
    x_lte_coeff = 1.0 / 12 * suggested_step ** 3

    if predict and len(pv_array) > 2 and pv_array[0][1] is not None and \
        pv_array[1][1] is not None and pv_array[2][1] is not None:
        A = np.zeros((len(pv_array[0]), len(pv_array[0])))
        A[0, :] = 0
        A[:, 0] = 1
        for row in range_type(1, A.shape[0]):
            for col in range_type(1, A.shape[0]):
                A[row, col] = (pv_array[row][0] - pv_array[0][0]) ** col
        Ainv = inv(A)

        predict_x = np.zeros(pv_array[0][1].shape)
        z = np.zeros((3, 1))
        for var in range_type(pv_array[0][1].shape[0]):
            for index in range(z.shape[0]):
                z[index, 0] = pv_array[index][1][var, 0]
            alpha = np.dot(Ainv, z)
            predict_x[var, 0] = alpha[2, 0] * suggested_step**2 + \
                                alpha[1, 0] * suggested_step + alpha[0, 0]
        predict_lte_coeff = -1.0 / 6.0 * suggested_step * \
                            (pv_array[0][0] + suggested_step - pv_array[1][0]) * \
                            (pv_array[0][0] + suggested_step - pv_array[2][0])
    else:
        predict_x, predict_lte_coeff = (None, None)
    return C1, C0, x_lte_coeff, predict_x, predict_lte_coeff




# -*- coding: iso-8859-1 -*-
# utilities.py
# Utilities file for Ahkab simulator
# Copyright 2006 Giuseppe Venturini

# This file is part of the ahkab simulator.
#
# Ahkab is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# Ahkab is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License v2
# along with ahkab.  If not, see <http://www.gnu.org/licenses/>.

"""
This module holds miscellaneous utility functions.

Module reference
################

"""

from __future__ import (unicode_literals, absolute_import,
                        division, print_function)

import collections
import os
import os.path
import operator

import numpy as np

from functools import wraps

from . import printing
from . import py3compat
from . import options

#: The machine epsilon, the upper bound on the relative error due to rounding in
#: floating point arithmetic.
EPS = np.finfo(float).eps


def expand_matrix(matrix, add_a_row=False, add_a_col=False):
    """Append a row and/or a column to the given matrix

    **Parameters:**

    matrix : ndarray
        The matrix to be manipulated.
    add_a_row : boolean, optional
        If set to ``True`` a row is appended to the supplied matrix.
    add_a_col : boolean
        If set to ``True`` a column is appended.

    **Returns:**

    matrix : ndarray
        A reference to the same matrix supplied.

    """
    n_row, n_col = matrix.shape
    if add_a_col:
        col = np.zeros((n_row, 1))
        matrix = np.concatenate((matrix, col), axis=1)
    if add_a_row:
        if add_a_col:
            n_col = n_col + 1
        row = np.zeros((1, n_col))
        matrix = np.concatenate((matrix, row), axis=0)
    return matrix

def set_submatrix(row, col, dest_matrix, source_matrix):
    """Copies a source matrix into another matrix

    row, col : ints
        The coordinates of the upper left corner in the destination matrix where
        the source matrix will be copied.
    dest_matrix : ndarray
        The matrix to be copied to.
    source_matrix : ndarray
        The matrix to be copied from.

    **Returns:**

    dest_matrix : ndarray
        A reference to the modified destination matrix.
    """
    ls = source_matrix.shape[0]
    cs = source_matrix.shape[1]
    dest_matrix[row:row+ls, col:col+cs] = source_matrix[:, :]
    return dest_matrix

def remove_row_and_col(matrix, rrow=0, rcol=0):
    """Removes a row and/or a column from a matrix

    **Parameters:**

    matrix : ndarray
        The matrix to be modified.
    rrow : int or None, optional
        The index of the row to be removed. If set to ``None``, no row
        will be removed. By default the first row is removed.
    rcol : int or None, optional
        The index of the row to be removed. If set to ``None``, no row
        will be removed. By default the first column is removed.

    .. note::

        No size checking is done.

    **Returns:**

    matrix : ndarray
        A reference to the modified matrix.
    """
    if rrow is not None and rcol is not None:
        return np.vstack((np.hstack((matrix[0:rrow, 0:rcol],
                                     matrix[0:rrow, rcol+1:])),
                          np.hstack((matrix[rrow+1:, 0:rcol],
                                     matrix[rrow+1:, rcol+1:]))
                          ))
    elif rrow is not None:
        return np.vstack((matrix[:rrow, :], matrix[rrow+1:, :]))
    elif rcol is not None:
        return np.hstack((matrix[:, :rcol], matrix[:, rcol+1:]))


def remove_row(matrix, rrow=0):
    """Removes a row from a matrix

    **Parameters:**

    matrix : ndarray
        The matrix to be modified.
    rrow : int or None, optional
        The index of the row to be removed. If set to ``None``, no row
        will be removed. By default the first row is removed.

    .. note::

        No size checking is done.

    **Returns:**

    matrix : ndarray
        A reference to the modified matrix.
    """
    return np.vstack((matrix[:rrow, :], matrix[rrow+1:, :]))


def check_file(filename):
    """Checks whether the supplied path refers to a valid file.

    **Parameters:**

    filename : string
        The file name.

    **Returns:**

    chk : boolean
        A value of ``True`` if ``filename`` is found and it is a file.

    :raises IOError: if no such file exists or if the supplied file is a directory.
    """
    filename = os.path.abspath(filename)
    if not os.path.exists(filename):
        raise IOError(filename + " not found.")
    elif not os.path.isfile(filename):
        raise IOError(filename + " is not a file.")
    return True

def tuplinator(alist):
    """Convert a list of lists (of lists...) to tuples"""
    if type(alist) is list:
        return tuple([tuplinator(i) for i in alist])
    else:
        return alist

class combinations:

    """This class is an iterator that returns all the k-combinations
    _without_repetition_ of the elements of the supplied list.

    Each combination is made of a subset of the list, consisting of k
    elements.

    **Parameters:**

    L : list
        The set from which the elements are taken.
    k : int
        The size of the subset, the number of elements to be taken
    """

    def __init__(self, L, k):
        self.L = L
        self.k = k
        self._sub_iter = None
        self._i = 0
        if len(self.L) < k:
            raise ValueError("The set has to be bigger than the subset.")
        if k <= 0:
            raise ValueError("The size of the subset has to be strictly " +
                             "positive.")

    def __iter__(self):
        return self

    def next(self):
        return self.__next__()

    def __next__(self):
        # It's recursive
        if self.k > 1:
            if self._sub_iter == None:
                self._sub_iter = combinations(self.L[self._i + 1:], self.k - 1)
            try:
                nxt = self._sub_iter.__next__()
                cur = self.L[self._i]
            except StopIteration:
                if self._i < len(self.L) - self.k:
                    self._i = self._i + 1
                    self._sub_iter = combinations(
                        self.L[self._i + 1:], self.k - 1)
                    return self.__next__()
                else:
                    raise StopIteration
        else:
            nxt = []
            if self._i < len(self.L):
                cur = self.L[self._i]
                self._i = self._i + 1
            else:
                raise StopIteration

        return [cur] + nxt


class log_axis_iterator:
    """This iterator provides the values for a base-10 logarithmic sweep.

    **Parameters:**

    min : float
        The minimum value, also the start point of the axis.
    max : float
        The maximum value, also the end point of the axis.
    points : int
        The number of points which will be used to discretize the ``max`` -
        ``min`` interval.

    Notice that, differently from numpy's ``logspace()``, the
    values are only computed at access time, and hence the
    memory footprint of the iterator is low.

    Start and end values are always included.
    """

    def __init__(self, min, max, points):
        self.inc = 10.**((np.log10(max) - np.log10(min))/(points - 1))
        self.max = max
        self.min = min
        self.index = 0
        self.current = min
        self.points = points

    def next(self):
        return self.__next__()

    def __next__(self):
        """Iterator method: get the next value
        """
        if self.index == 0:
            ret = self.current
        elif self.index < self.points:
            self.current = self.current * self.inc
            ret = self.current
        else:
            raise StopIteration
        self.index = self.index + 1
        return ret

    def __getitem__(self, i):
        """Iterator method: get a particular value (n. i)
        """
        if i == 0:
            ret = self.min
        elif i < self.points:
            ret = self.min * self.inc**i
        else:
            ret = None
        return ret

    def __iter__(self):
        """Required iterator method.
        """
        return self


class lin_axis_iterator:
    """This iterator provides the values for a linear sweep.

    **Parameters:**

    min : float
        The minimum value, also the start point of the axis.
    max : float
        The maximum value, also the end point of the axis.
    num : int
        The number of samples to generate. In general, this should be greater than 1.
        A value of 1 is accepted only if ``min == max``, in which case, only one
        value is returned by the iterator: ``min``. 

    Start and end points are always included.

    Notice that, differently from numpy's ``linspace()``, the
    values are only computed at access time, and hence the
    memory footprint of the iterator is low.

    :raises ValueError: if the number ``points`` is either negative or does not
    respect the conditions above.

    """

    def __init__(self, min, max, points):
        if points < 2 and min != max:
            raise ValueError('Linear iterator from %d to %d with %d points.'%
                             (min, max, points))
        if points < 1:
            raise ValueError('Linear iterator from %d to %d with %d points.'%
                             (min, max, points))
        if points > 1:
            self.inc = (max - min) / (points - 1)
        elif points == 1 and max == min:
            # sometimes, they ask for this. They expect to get back [max]
            self.inc = 0
        self.max = max
        self.min = min
        self.index = 0
        self.current = min
        self.points = points

    def next(self):
        return self.__next__()

    def __next__(self):
        """Iterator method: get the next value
        """
        if self.index == 0:
            pass  # return min
        elif self.index < self.points:
            self.current = self.current + self.inc
        else:
            raise StopIteration
        ret = self.current
        self.index = self.index + 1
        return ret

    def __getitem__(self, i):
        """Iterator method: get a particular value (n. i)
        """
        if i < self.points:
            ret = self.min + self.inc * i
        else:
            ret = None
        return ret

    def __iter__(self):
        """Required iterator method.
        """
        return self


def Celsius2Kelvin(cel):
    """Convert Celsius degrees to Kelvin
    """
    return cel + 273.15


def Kelvin2Celsius(kel):
    """Convert Kelvin degrees to Celsius
    """
    return kel - 273.15

def convergence_check(x, dx, residuum, nv_minus_one, debug=False):
    """Perform a convergence check

    **Parameters:**

    x : array-like
        The results to be checked.
    dx : array-like
        The last increment from a Newton-Rhapson iteration, solving
        ``F(x) = 0``.
    residuum : array-like
        The remaining error, ie ``F(x) = residdum``
    nv_minus_one : int
        Number of voltage variables in x. If ``nv_minus_one`` is equal to
        ``n``, it means ``x[:n]`` are all voltage variables.
    debug : boolean, optional
        Whether extra information is needed for debug purposes. Defaults to
        ``False``.

    **Returns:**

    chk : boolean
        Whether the check was passed or not. ``True`` means 'convergence!'.
    rbn : ndarray
        The convergence check results by node, if ``debug`` was set to ``True``,
        else ``None``.
    """
    if not hasattr(x, 'shape'):
        x = np.array(x)
        dx = np.array(dx)
        residuum = np.array(residuum)
    vcheck, vresults = voltage_convergence_check(x[:nv_minus_one, 0],
                                                 dx[:nv_minus_one, 0],
                                                 residuum[:nv_minus_one, 0])
    icheck, iresults = current_convergence_check(x[nv_minus_one:],
                                                 dx[nv_minus_one:],
                                                 residuum[nv_minus_one:])
    return vcheck and icheck, vresults + iresults


def voltage_convergence_check(x, dx, residuum, debug=False):
    """Perform a convergence check for voltage variables

    **Parameters:**

    x : array-like
        The results to be checked.
    dx : array-like
        The last increment from a Newton-Rhapson iteration, solving
        ``F(x) = 0``.
    residuum : array-like
        The remaining error, ie ``F(x) = residdum``
    debug : boolean, optional
        Whether extra information is needed for debug purposes. Defaults to
        ``False``.

    **Returns:**

    chk : boolean
        Whether the check was passed or not. ``True`` means 'convergence!'.
    rbn : ndarray
        The convergence check results by node, if ``debug`` was set to ``True``,
        else ``None``.
    """
    return custom_convergence_check(x, dx, residuum, er=options.ver,
                                    ea=options.vea, eresiduum=options.iea,
                                    debug=debug)


def current_convergence_check(x, dx, residuum, debug=False):
    """Perform a convergence check for current variables

    **Parameters:**

    x : array-like
        The results to be checked.
    dx : array-like
        The last increment from a Newton-Rhapson iteration, solving
        ``F(x) = 0``.
    residuum : array-like
        The remaining error, ie ``F(x) = residdum``
    debug : boolean, optional
        Whether extra information is needed for debug purposes. Defaults to
        ``False``.

    **Returns:**

    chk : boolean
        Whether the check was passed or not. ``True`` means 'convergence!'.
    rbn : ndarray
        The convergence check results by node, if ``debug`` was set to ``True``,
        else ``None``.
    """
    return custom_convergence_check(x, dx, residuum, er=options.ier,
                                    ea=options.iea, eresiduum=options.vea,
                                    debug=debug)


def custom_convergence_check(x, dx, residuum, er, ea, eresiduum, debug=False):
    """Perform a custom convergence check

    **Parameters:**

    x : array-like
        The results to be checked.
    dx : array-like
        The last increment from a Newton-Rhapson iteration, solving
        ``F(x) = 0``.
    residuum : array-like
        The remaining error, ie ``F(x) = residdum``
    ea : float
        The value to be employed for the absolute error.
    er : float
        The value for the relative error to be employed.
    eresiduum : float
        The maximum allowed error for the residuum (left over error).
    debug : boolean, optional
        Whether extra information is needed for debug purposes. Defaults to
        ``False``.

    **Returns:**

    chk : boolean
        Whether the check was passed or not. ``True`` means 'convergence!'.
    rbn : ndarray
        The convergence check results by node, if ``debug`` was set to ``True``,
        else ``None``.
    """
    all_check_results = []
    if not hasattr(x, 'shape'):
        x = np.array(x)
        dx = np.array(dx)
        residuum = np.array(residuum)
    if x.shape[0]:
        if not debug:
            ret = np.allclose(x, x + dx, rtol=er, atol=ea) and \
                  np.allclose(residuum, np.zeros(residuum.shape),
                              atol=eresiduum, rtol=0)
        else:
            for i in range(x.shape[0]):
                if np.abs(dx[i, 0]) < er*np.abs(x[i, 0]) + ea and \
                   np.abs(residuum[i, 0]) < eresiduum:
                    all_check_results.append(True)
                else:
                    all_check_results.append(False)
                if not all_check_results[-1]:
                    break

            ret = not (False in all_check_results)
    else:
        # We get here when there's no variable to be checked. This is because
        # there aren't variables of this type.  Eg. the circuit has no voltage
        # sources nor voltage defined elements. In this case, the actual check
        # is done only by current_convergence_check, voltage_convergence_check
        # always returns True.
        ret = True

    return ret, all_check_results

def check_step_and_points(step=None, points=None, period=None,
                          default_points=100):
    """Sets consistently the step size and the number of points

    The calculation is done according to the given period.

    **Parameters:**

    step : scalar, optional
        The discretization step.
    points : int, optional
        The number of points to be used in the discretization.
    period : scalar, optional
        The length of the interval to be discretized. Not setting
        this parameter will result in a ``ValueError``.
    default_points : int, optional
        The default number of points.

    **Returns:**

    (points, step) : tuple
        The adjusted number of points and step value.
    """

    if step is None and points is None:
        printing.print_warning("Neither step nor n. of points set. Using %d points." % default_points)
        points = default_points
    elif step is not None and points is not None:
        printing.print_warning("Both step and # of points set. Using step = %f." % step)
        points = None

    if points:
        step = float(period)/(points - 1)
    else:
        points = float(period)/step
        if points % 1 != 0:
            step = step + (step * (points % 1)) / int(points)
            points = int(float(period)/step)
            printing.print_warning("adapted step is %g" % (step,))
        else:
            points = int(points)
        # 0 - N where xN is in reality the first point of the second period!!
        points = points + 1

    return int(points), step

def check_circuit(circ):
    """Performs some easy sanity checks.

    Checks performed:

    * Has the circuit more than one node?
    * Has the circuit a connection to ground?
    * Has the circuit more than two elements?
    * Are there no two elements with the same ``part_id``?

    **Parameters:**

    circ : circuit instance
        The circuit to be checked.

    **Returns:**

    chk : boolean
        The logical ``and()`` of the answer to the above questions.
    msg : string
        A message describing the error, if any.
    """

    if circ.get_nodes_number() < 2:
        test_passed = False
        reason = "the circuit has less than two nodes."
    elif not 0 in circ.nodes_dict:
        test_passed = False
        reason = "the circuit has no ref. Quitting."
    elif len(circ) < 2:
        test_passed = False
        reason = "the circuit has less than two elements."
    elif circ.has_duplicate_elem():
        test_passed = False
        reason = "duplicate elements found (check the names!)"
    else:
        test_passed = True
        reason = ""

    return test_passed, reason


def check_ground_paths(mna, circ, reduced_mna=True, verbose=3):
    """Checks that every node has a DC path to ground

    The path to ground might be through non-linear elements.

    .. note::

        * This does not ensure that the circuit will have a DC solution.
        * A node without DC path to ground would be rescued (likely) by GMIN so
          (for the time being at least) we do *not* halt the execution.
        * Also, two series capacitors always fail this check (GMIN saves us)

    Bottom line: if there is no DC path to ground, there is probably a
    mistake in the netlist. Print a warning.

    **Returns:**

    chk : boolean
        A boolean set to true if there is a DC path to ground from all nodes
        in the circuit.
    """
    test_passed = True
    if reduced_mna:
        # reduced_correction
        r_c = 1
    else:
        r_c = 0
    to_be_checked_for_nonlinear_paths = []
    for node in iter(circ.nodes_dict.keys()):
        if node == 0:
            continue
            # ground
        if type(node) != int:
            # an ext handle
            continue
        if mna[node - r_c, node - r_c] == 0 and \
           not mna[node - r_c, circ.get_nodes_number() - r_c:].any():
            to_be_checked_for_nonlinear_paths.append(node)
    for node in to_be_checked_for_nonlinear_paths:
        node_is_nl_op = False
        for elem in circ:
            if not elem.is_nonlinear:
                continue
            ops = elem.get_output_ports()
            for op in ops:
                if op.count(node):
                    node_is_nl_op = True
        if not node_is_nl_op:
            if verbose:
                printing.print_warning(
                    "No path to ground from node " + circ.nodes_dict[node])
            test_passed = False
    return test_passed

def memoize(f):
    """Memoization decorator

    **Parameters:**

    f : function
        The function to apply memoization to.

    **Returns:**

    fm : function
        The function with added memoization.

    **Implementation:**

    Originally from `this post
    <https://code.activestate.com/recipes/578231-probably-the-fastest-memoization-decorator-in-the-/#c4>`_,
    it has been modified to provide a cache of size ``options.cache_len``.

    .. note::

        The size of the cache is per model instance and per function. If you
        have one model, shared by several elements, you probably prefer to have
        a big cache.

    """
    class memodict(collections.OrderedDict):
        __slots__ = ()
        @wraps(f)
        def __getitem__(self, *key):
            return dict.__getitem__(self, key)
        def __missing__(self, key):
            ret = self[key] = f(*key)
            # set options.cache_len to None to disable any size limit.
            if options.cache_len is not None and len(self) > options.cache_len:
                self.popitem() #FIFO pop
            return ret
    return memodict().__getitem__





# -*- coding: iso-8859-1 -*-
# ac.py
# AC analysis module
# Copyright 2010 Giuseppe Venturini
# This file is part of the ahkab simulator.
#
# Ahkab is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# Ahkab is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License v2
# along with ahkab.  If not, see <http://www.gnu.org/licenses/>.

"""This module contains the methods required to perform an AC analysis.

.. note::

    Typically, the user does not need to call the functions in this module
    directly, instead, we recommend defining an AC analysis object through the
    convenience method :func:`ahkab.ahkab.new_ac` and then running it calling
    :func:`ahkab.ahkab.run`.

Overview of AC simulations
--------------------------

The AC simulation problem
'''''''''''''''''''''''''

Our AC analysis problem can be written as:

.. math::

    MNA\\ x + AC(\\omega )\\ x + J x + N_{ac}(\\omega) = 0

We need:

    1. the Modified Nodal Analysis matrix :math:`MNA`,
    2. the :math:`AC` matrix, holding the frequency dependent parts,
    3. :math:`J`, the Jacobian matrix from the linearized non-linear elements,
    4. :math:`N_{ac}`, the AC sources contribution.

An Operating Point (OP) has to be computed first if there is any non-linear
device in the circuit to perform the linearization.

When all the matrices are available, it is possible to solve the system
for the frequency values specified by the user, providing the resulting
matrix is not singular (and possibly well conditioned).

Building the AC matrix
''''''''''''''''''''''

It's easy to set up the voltage lines, since line 2 refers to
node 2, etc...

A capacitor between two example nodes ``n1`` and ``n2`` introduces the
following elements:

.. math::

    \\mathrm{(KCL\\ node\\ n1)}\\qquad +j\\omega C\\ V(n1) - j\\omega C V(n2) + ... = ...

    \\mathrm{(KCL\\ node\\ n2)}\\qquad -j\\omega C\\ V(n1) + j\\omega C V(n2) + ... = ...

Inductors generate, together with voltage sources, Current-Controlled Voltage
sources (CCVS), Voltage-Controlled Voltage Sources (VCVS), an
additional line in the :math:`MNA` matrix, and hence in :math:`AC` too.
In fact, the current flowing through the device gets added to the unknowns
vector, :math:`x`.

For example, in the case of an inductors, we have:

.. math::

    \\mathrm{(KVL\\ over\\ n1\\ and\\ n2)}\\qquad V(n1) - V(n2) - j\\omega L\\ I(\\mathrm{inductor}) = 0

To understand on which line is the KVL line for an inductor, we use the
*order* of the elements in :mod:`ahkab.circuit`:

* first are assembled all the voltage rows,
* then the current rows, in the same order in which the elements that introduce
  them are found in :class:`ahkab.circuit.Circuit`.

Solving
'''''''

For each angular frequency :math:`\\omega`, the simulator solves the matrix
equation described.

Since the equation is linear, solving is performed with a single matrix
inversion and multiplication for each step.

Module reference
----------------

"""

from __future__ import (unicode_literals, absolute_import,
                        division, print_function)

import numpy as np

from . import (circuit, dc_analysis, devices, options, printing, results,
               utilities)

specs = {'ac': {'tokens': ({
                           'label': 'type',
                           'pos': 0,
                           'type': str,
                           'needed': False,
                           'dest': 'sweep_type',
                           'default': options.ac_log_step
                           },
                          {
                           'label': 'nsteps',
                           'pos': 1,
                           'type': float,
                           'needed': True,
                           'dest': 'points',
                           'default': None
                          },
                          {
                           'label': 'start',
                           'pos': 2,
                           'type': float,
                           'needed': True,
                           'dest': 'start',
                           'default': None
                          },
                          {
                           'label': 'stop',
                           'pos': 3,
                           'type': float,
                           'needed': True,
                           'dest': 'stop',
                           'default': None
                          })
               }
        }


def ac_analysis(circ, start, points, stop, sweep_type=None,
                x0=None, mna=None, AC=None, Nac=None, J=None,
                outfile="stdout", verbose=3):
    """Performs an AC analysis.

    **Parameters:**

    circ : Circuit instance
        The circuit to be simulated.

    start : float
        The start frequency for the AC analysis, in Hz.

    points : float,
        The number of points to be used to discretize the
        ``[start, stop]`` interval.

    stop : float
        The stop frequency, in Hz.

    sweep_type : string, optional
        Either ``options.ac_log_step`` (ie ``'LOG'``) or ``options.ac_lin_step``
        (ie ``'LIN'``), defaults to ``options.ac_log_step``, resulting in a
        logarithmic sweep.

    x0 : OP results instance, optional
        The linearization point. If not set, it will be computed
        running an OP analysis.

    mna, AC, Nax, J : ndarrays, optional
        The matrices to perform the analysis. They will be computed
        if not supplied.

    outfile : string, optional
        The name of the file where the results will be written.
        The suffix ``'.ac'`` is automatically added at the end of the string to
        prevent different analyses from overwriting each-other's results. Set to
        ``'stdout'`` to write to the standard output.  If unset, or set to
        ``None``, defaults to the standard output.

    verbose : int, optional
        The verbosity level, from 0 (silent) to 6 (debug).

    **Returns:**

    ACresult : AC solution
        The AC analysis results.

    :raises ValueError: if the parameters are out of their valid range.

    :raises RuntimeError: if the circuit is non-linear and can't be linearized.
    """

    if outfile == 'stdout':
        verbose = 0

    # check step/start/stop parameters
    if start == 0:
        raise ValueError("AC analysis has start frequency = 0")
    if start > stop:
        raise ValueError("AC analysis has start > stop")
    if points < 2 and not start == stop:
        raise ValueError("AC analysis has number of points < 2 & start != stop")
    if sweep_type.upper() == options.ac_log_step or sweep_type is None:
        omega_iter = utilities.log_axis_iterator(2*np.pi*start, 2*np.pi*stop, points)
    elif sweep_type.upper() == options.ac_lin_step:
        omega_iter = utilities.lin_axis_iterator(2*np.pi*start, 2*np.pi*stop, points)
    else:
        raise ValueError("Unknown sweep type %s" % sweep_type)

    tmpstr = "Vea =", options.vea, "Ver =", options.ver, "Iea =", options.iea, "Ier =", \
        options.ier, "max_ac_nr_iter =", options.ac_max_nr_iter
    printing.print_info_line((tmpstr, 5), verbose)
    del tmpstr

    printing.print_info_line(("Starting AC analysis: ", 1), verbose)
    tmpstr = "w: start = %g Hz, stop = %g Hz, %d points" % (start, stop, points)
    printing.print_info_line((tmpstr, 3), verbose)
    del tmpstr

    # It's a good idea to call AC with prebuilt MNA matrix if the circuit is
    # big
    if mna is None:
        mna, N = dc_analysis.generate_mna_and_N(circ, verbose=verbose)
        del N
        mna = utilities.remove_row_and_col(mna)
    if Nac is None:
        Nac = _generate_Nac(circ)
        Nac = utilities.remove_row(Nac, rrow=0)
    if AC is None:
        AC = _generate_AC(circ, [mna.shape[0], mna.shape[0]])
        AC = utilities.remove_row_and_col(AC)

    if circ.is_nonlinear():
        if J is not None:
            pass
            # we used the supplied linearization matrix
        else:
            if x0 is None:
                printing.print_info_line(("Starting OP analysis to get a " +
                                          "linearization point...", 3), verbose,
                                         print_nl=False)
                # silent OP
                x0 = dc_analysis.op_analysis(circ, verbose=0)
                if x0 is None:  # still! Then op_analysis has failed!
                    printing.print_info_line(("failed.", 3), verbose)
                    raise RuntimeError("OP analysis failed, no " +
                                       "linearization point available.")
                else:
                    printing.print_info_line(("done.", 3), verbose)
            printing.print_info_line(
                ("Linearization point (xop):", 5), verbose)
            if verbose > 4:
                x0.print_short()
            printing.print_info_line(("Linearizing the circuit...", 5), verbose,
                                     print_nl=False)
            J = _generate_J(xop=x0.asarray(), circ=circ,
                            reduced_mna_size=mna.shape[0])
            printing.print_info_line((" done.", 5), verbose)
            # we have J, continue
    else:  # not circ.is_nonlinear()
        # no J matrix is required.
        J = 0

    printing.print_info_line(("MNA (reduced):", 5), verbose)
    printing.print_info_line((mna, 5), verbose)
    printing.print_info_line(("AC (reduced):", 5), verbose)
    printing.print_info_line((AC, 5), verbose)
    printing.print_info_line(("J (reduced):", 5), verbose)
    printing.print_info_line((J, 5), verbose)
    printing.print_info_line(("Nac (reduced):", 5), verbose)
    printing.print_info_line((Nac, 5), verbose)

    sol = results.ac_solution(circ, start=start, stop=stop, points=points,
                              stype=sweep_type, op=x0, outfile=outfile)

    # setup the initial values to start the iteration:
    j = np.complex('j')

    Gmin_matrix = dc_analysis.build_gmin_matrix(
        circ, options.gmin, mna.shape[0], verbose)

    iter_n = 0  # contatore d'iterazione
    printing.print_info_line(("Solving... ", 3), verbose, print_nl=False)

    x = x0
    for omega in omega_iter:
        x, _, solved, _ = dc_analysis.dc_solve(
            mna=(mna + np.multiply(j * omega, AC) + J),
            Ndc = Nac,
            Ntran = 0,
            circ = circuit.Circuit(
                title="Dummy circuit for AC", filename=None),
            Gmin = Gmin_matrix,
            x0 = x,
            time = None,
            locked_nodes = None,
            MAXIT = options.ac_max_nr_iter,
            skip_Tt = True,
            verbose = 0)
        if solved:
            iter_n = iter_n + 1
            # hooray!
            sol.add_line(omega/np.pi/2, x)
        else:
            break

    if solved:
        printing.print_info_line(("done.", 1), verbose)
        ret_value = sol
    else:
        printing.print_info_line(("failed.", 1), verbose)
        ret_value = None

    return ret_value


def _generate_AC(circ, shape):
    """Generates the AC coefficients matrix.

    **Parameters:**

    circ : Circuit instance
        The circuit instance for which the :math:`AC` matrix will be generated.

    shape : int
        The reduced MNA size.

    **Returns:**

    AC : ndarray
        The *unreduced* (full size) :math:`AC` matrix.

    """
    AC = np.zeros((shape[0] + 1, shape[1] + 1))
    nv = circ.get_nodes_number()  # - 1
    i_eq = 0  # each time we find a vsource or vcvs or ccvs, we'll add one to this.
    for elem in circ:
        if circuit.is_elem_voltage_defined(elem) and not isinstance(elem, devices.Inductor):
            i_eq = i_eq + 1
        elif isinstance(elem, devices.Capacitor):
            n1 = elem.n1
            n2 = elem.n2
            AC[n1, n1] = AC[n1, n1] + elem.value
            AC[n1, n2] = AC[n1, n2] - elem.value
            AC[n2, n2] = AC[n2, n2] + elem.value
            AC[n2, n1] = AC[n2, n1] - elem.value
        elif isinstance(elem, devices.Inductor):
            AC[nv + i_eq, nv + i_eq] = -1 * elem.value
            if len(elem.coupling_devices):
                for cd in elem.coupling_devices:
                    # get `part_id` of the other inductor (eg. "L32")
                    other_id_wdescr = cd.get_other_inductor(elem.part_id)
                    # find its index to know which column corresponds to its
                    # current
                    other_index = circ.find_vde_index(
                        other_id_wdescr, verbose=0)
                    # add the term.
                    AC[nv + i_eq, nv + other_index] += -1 * cd.M
            i_eq = i_eq + 1

    if options.cmin > 0:
        cmin_mat = np.eye(shape[0] + 1 - i_eq)
        cmin_mat[0, 1:] = 1
        cmin_mat[1:, 0] = 1
        cmin_mat[0, 0] = cmin_mat.shape[0] - 1
        if i_eq:
            AC[:-i_eq, :-i_eq] += options.cmin * cmin_mat
        else:
            AC += options.cmin * cmin_mat

    return AC


def _generate_Nac(circ):
    """Generate the vector holding the contribution of AC sources.

    **Parameters:**

    circ : Circuit instance
        The circuit instance for which the :math:`N_{ac}` matrix will be
        generated.

    **Returns:**

    Nac : ndarray
        The constant term :math:`N_{ac}`.
    """
    n_of_nodes = circ.get_nodes_number()
    Nac = np.zeros((n_of_nodes, 1), dtype=complex)
    j = np.complex('j')
    # process `ISource`s
    for elem in circ:
        if isinstance(elem, devices.ISource) and elem.abs_ac is not None:
            # convenzione normale!
            Nac[elem.n1, 0] = Nac[elem.n1, 0] + \
                elem.abs_ac * np.exp(j * elem.arg_ac)
            Nac[elem.n2, 0] = Nac[elem.n2, 0] - \
                elem.abs_ac * np.exp(j * elem.arg_ac)
    # process vsources
    # for each vsource, introduce a new variable: the current flowing through it.
    # then we introduce a KVL equation to be able to solve the circuit
    for elem in circ:
        if circuit.is_elem_voltage_defined(elem):
            index = Nac.shape[0]
            Nac = utilities.expand_matrix(Nac, add_a_row=True, add_a_col=False)
            if isinstance(elem, devices.VSource) and elem.abs_ac is not None:
                Nac[index, 0] = -1.0 * elem.abs_ac * np.exp(j * elem.arg_ac)
    return Nac


def _generate_J(xop, circ, reduced_mna_size):
    """Build the linearized matrix :math:`J`

    **Parameters:**

    xop : ndarray
        The linearization point, as a ``numpy`` ndarray.
    circ : Circuit instance
        The circuit for which :math:`J` is to be generated.
    reduced_mna_size : int
        The size of the (square) Modified Nodal Analysis Matrix, after
        reduction.

    **Returns:**

    J : ndarray of size ``(reduced_mna_size, reduced_mna_size)``
        The reduced Jacobian :math:`J`.

    """
    # setup J
    J = np.zeros((reduced_mna_size, reduced_mna_size))
    Tlin = np.zeros((reduced_mna_size, 1))
    for elem in circ:
        if elem.is_nonlinear:
            dc_analysis._update_J_and_Tx(J, Tlin, xop, elem, time=None)
    # del Tlin # not needed! **DC**!
    return J



# -*- coding: iso-8859-1 -*-
# plotting.py
# Module to plot the simulation data (through matplotlib)
# Copyright 2009-2013 Giuseppe Venturini

# This file is part of the ahkab simulator.
#
# Ahkab is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# Ahkab is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License v2
# along with ahkab.  If not, see <http://www.gnu.org/licenses/>.

"""
This module offers the functions needed to plot the results
of a simulation.

It is only functional if `matplotlib <http://matplotlib.org/>`_
is installed.

Module reference
''''''''''''''''

"""

from __future__ import (unicode_literals, absolute_import,
                        division, print_function)

import re
import numpy as np

from . import printing
from . import options
from . import py3compat

try:
    import pylab
except ImportError:
    # no matplotlib on a platform that supports it
    if not py3compat.PYPY:
        printing.print_warning('Matplotlib not found.')

def _split_netlist_label(labels_string):
    labels_string = labels_string.strip().upper()
    ret_labels = []
    p = re.compile(r'V\s*\(\s*(\w*)\s*,\s*(\w*)\s*\)', re.IGNORECASE)
    labels_list = p.findall(labels_string)
    for i in range(len(labels_list)):
        l2 = "V" + labels_list[i][0]
        l1 = "V" + labels_list[i][1]
        if l1 != 'V0':
            ret_labels.append((l2, l1))
        else:
            ret_labels.append((l2,))
    p = re.compile(r'V\s*\(\s*(\w*)\s*\)', re.IGNORECASE)
    labels_list = p.findall(labels_string)
    for i in range(len(labels_list)):
        l2 = "V" + labels_list[i]
        l1 = None
        ret_labels.append((l2, l1))
    p = re.compile(r'I\s*\(\s*(\w*)\s*\)', re.IGNORECASE)
    labels_list = p.findall(labels_string)
    for i in range(len(labels_list)):
        l2 = "I(" + labels_list[i] + ")"
        l1 = None
        ret_labels.append((l2, l1))
    if len(ret_labels) == 0:
        raise ValueError("Unrecognized plot labels: " + labels_string)
    return ret_labels


def _setup_plot(fig, title, xvu, yvu, log=False, xlog=False, ylog=False):
    """Setup the figure for plotting. 

    **Parameters:**

    fig : figure
        A matplotlib figure
    title : string
        The plot title:
    xvu : tuple
        A tuple defined as ``xvu = (xvar, xunit)``, where ``xvar`` is the
        x-axis variable label (str) and ``xunit`` is its unit (str too).
    yvu : list of tuples
        defined as yvu += [
        Each tuple defined as ``(yvar, yunit)``, where ``yvarN`` is the
        y-axis variable label (str) and ``yunit`` is its unit (str too).
    log : bool, optional
        Whether to set all scales to ``log``.
    xlog : bool, optional
        Whether to set the x-axis scale to ``log``.
    ylog : bool, optional
        Whether to set the y-axis scale to ``log``.

    """
    # set the currently active figure to fig
    pylab.figure(fig.number)
    # xvar, xunit = xvu
    pylab.title(title.upper())
    ax = pylab.gca()

    ax.spines['right'].set_color('none')
    ax.spines['top'].set_color('none')
    ax.xaxis.set_ticks_position('bottom')
    ax.yaxis.set_ticks_position('left')

    ax.xaxis.grid(False)
    ax.yaxis.grid(False)

    if log or xlog:
        ax.set_xscale('log')
    pylab.xlabel('%s [%s]' % xvu)
    yunits = []
    yinitials = []
    for yv, yu in yvu:
        yv = yv[:].replace('|', "")
        if yu not in yunits:
            yunits.append(yu)
            yinitials.append(yv[0])
    ylabel = ""
    for yi, yu in zip(yinitials, yunits):
        ylabel += "%s [%s] , " % (yi, yu)
    ylabel = ylabel[:-3]
    pylab.ylabel(ylabel)

    if log or ylog:
        ax.set_yscale('log')
    # fig.tight_layout()


def save_figure(filename, fig=None):
    """Apply the figure options for saving and then save the supplied
    figure to ``filename``.
    
    The format of the output figure is set by ``options.plotting_outtype``.

    **Parameters:**

    filename : string
        The output filename.
    fig : figure object, optional
        The figure to be saved.

    **Returns:**
    
    ``None``.
    """
    if fig is None:
        fig = pylab.gcf()
    fig.set_size_inches(*options.plotting_save_figsize)
    pylab.savefig(filename, dpi=100, bbox_inches='tight',
                  format=options.plotting_outtype, pad=0.1)
    fig.set_size_inches(*options.plotting_display_figsize)

def _data_abs_arg_pass(res, label):
    # extract abs / phase if needed or pass the data
    if label[0] == label[-1] == '|':
        data = np.absolute(res[label[1:-1]])
        units = res.units[label[1:-1]]
    elif label[0:4] == 'arg(' and label[-1] == ')':
        data = np.angle(res[label[4:-1]], deg=options.ac_phase_in_deg)
        units = res.units[label[4:-1]]
    else:
        data = res[label]
        units = res.units[label]
    return data, units

def plot_results(title, y2y1_list, results, outfilename=None):
    """Plot the results.

    **Parameters:**

    title : string
        The plot title
    y2y1_list : list
        A list of tuples. Each tuple has to be in the format ``(y2, y1)``.
        Each member of the tuple has to be a valid identifier. You can 
        check the possible voltage and current identifiers in the 
        result set calling ``res.keys()``, where ``res`` is a solution
        object.
    result : solution object or derivate
        The results to be plotted.
    outfilename : string, optional
        The filename of the output file. If left unset, the plot will
        not be written to disk. The format is set through
        ``options.plotting_outtype``.

    **Returns:**

    ``None``.
    """
    if results is None:
        printing.print_warning("No results available for plotting. Skipping.")
        return
    fig = pylab.figure(figsize=options.plotting_display_figsize)
    analysis = results.sol_type.upper()
    gdata = []
    x, xlabel = results.get_x(), results.get_xlabel()
    xunit = results.units[xlabel]
    yvu = []

    for y2label, y1label in y2y1_list:
        if y1label is not None and y1label != '':
            try:
                data1, _ = _data_abs_arg_pass(results, y1label)
            except ValueError as e:
                printing.print_warning(str(e) + " " + y1label)
                continue
            line_label = y2label + "-" + y1label
        else:
            line_label = y2label
            data1 = 0
        try:
            data2, units = _data_abs_arg_pass(results, y2label)
        except ValueError as e:
            printing.print_warning(str(e) + " " + y2label)
            continue
        yvu += [(line_label, units)]
        gdata.append((data2 - data1, line_label))

    if xlabel == 'f':
        xlog = True
    else:
        xlog = False
    _setup_plot(fig, title, (xlabel, xunit), yvu, xlog=xlog)
    ms = 7./(1. + max(np.log(len(x)/100.), 0))
    ms = ms if ms > 2 else 0.

    pylab.hold(True)
    ymax, ymin = None, None
    for y, label in gdata:
        [line] = pylab.plot(
            x, y, options.plotting_style, label=label +
            " (" + analysis + ")", ms=ms,
            mfc='w', lw=options.plotting_lw, mew=options.plotting_lw)
        line.set_mec(line.get_color())  # nice empty circles
        ymax = y.max() if ymax is None or y.max() > ymax else ymax
        ymin = y.min() if ymin is None or y.min() < ymin else ymin
    pylab.xlim((x.min(), x.max()))
    pylab.ylim((ymin - (ymax - ymin) * .01, ymax + (ymax - ymin) * .01))
    pylab.grid(True)
    pylab.hold(False)
    pylab.legend()

    if outfilename is not None and options.plotting_outtype is not None:
        save_figure(outfilename, fig)
    return


def show_plots():
    """See the fruit of your work!"""
    pylab.show()



# -*- coding: iso-8859-1 -*-
# custom.py
# custom model
# Copyright 2015 Giuseppe Venturini

# This file is part of the ahkab simulator.
#
# Ahkab is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# Ahkab is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License v2
# along with ahkab.  If not, see <http://www.gnu.org/licenses/>.

"""
This module contains a custom element and its model class.

.. image:: images/elem/custom.svg

"""

#
#             +---+
#      n1 o---|   |---o n2
#             | G |
#      n3 o---|   |---o n4
#          .  |   |  .
#          .  |   |  .
#          .  |   |  .
#  n(N-1) o---|   |---o nN
#             +---+
#

from __future__ import (unicode_literals, absolute_import,
                        division, print_function)

import numpy as np


from . import constants
from . import utilities
from . import options

damping_factor = 4.

class custom(devices.Component):
    """A custom element.

    **Parameters:**

    n1, n2 : string
        The custom anode and cathode.
    model : model instance
        The custom model providing the mathemathical modeling.
    ic : float
        The custom initial voltage condition for transient analysis
        (ie :math:`V_D = V(n_1) - V(n_2)` at :math:`t = 0`).
    off : bool
         Whether the custom should be initially assumed to be off when
         computing an OP.

    The other are the physical parameters reported in the following table:

    """

    def __init__(self, part_id, nodes, model, **parameters):
        super(ChildB, self).__init__(**parameters)
        class _dev_class(object):
            pass
        self.device = _dev_class()
        self.device.AREA = AREA if AREA is not None else 1.0
        self.device.T = T
        self.device.last_vd = .425
        self.n1 = n1
        self.n2 = n2
        self.ports = ((self.n1, self.n2),)
        self.model = model
        if self.device.T is None:
            self.device.T = constants.T

    def _get_T(self):
        return self.device.T

    def set_temperature(self, T):
        """Set the operating temperature IN KELVIN degrees
        """
        # this automatically makes the memoization cache obsolete. self.device
        # is in fact passed as one of the arguments, hashed and stored:
        # if it changes, the old cache will return misses.
        self.device.T = T

    def __str__(self):
        rep = "%s area=%g T=%g" % (
            self.model.name, self.device.AREA, self._get_T())
        if self.ic is not None:
            rep = rep + " ic=" + str(self.ic)
        elif self.off:
            rep += " off"
        return rep

    def get_output_ports(self):
        return self.ports

    def get_drive_ports(self, op):
        if not op == 0:
            raise ValueError('custom %s has no output port %d' %
                             (self.part_id, op))
        return self.ports

    def istamp(self, ports_v, time=0, reduced=True):
        """Get the current matrix

        A matrix corresponding to the current flowing in the element
        with the voltages applied as specified in the ``ports_v`` vector.

        **Parameters:**

        ports_v : list
            A list in the form: [voltage_across_port0, voltage_across_port1, ...]
        time: float
            the simulation time at which the evaluation is performed.
            It has no effect here. Set it to None during DC analysis.

        """
        v = ports_v[0]
        i = self.model.get_i(self.model, v, self.device)
        istamp = np.array((i, -i), dtype=np.float64)
        indices = ((self.n1 - 1*reduced, self.n2 - 1*reduced), (0, 0))
        if reduced:
            delete_i = [pos for pos, ix in enumerate(indices[0]) if ix == -1]
            istamp = np.delete(istamp, delete_i, axis=0)
            indices = tuple(zip(*[(ix, j) for ix, j in zip(*indices) if ix != -1]))
        return indices, istamp

    def i(self, op_index, ports_v, time=0):  # with gmin added
        v = ports_v[0]
        i = self.model.get_i(self.model, v, self.device)
        return i

    def gstamp(self, ports_v, time=0, reduced=True):
        """Returns the differential (trans)conductance wrt the port specified by port_index
        when the element has the voltages specified in ports_v across its ports,
        at (simulation) time.

        ports_v: a list in the form: [voltage_across_port0, voltage_across_port1, ...]
        port_index: an integer, 0 <= port_index < len(self.get_ports())
        time: the simulation time at which the evaluation is performed. Set it to
        None during DC analysis.
        """
        indices = ([self.n1 - 1]*2 + [self.n2 - 1]*2,
                   [self.n1 - 1, self.n2 - 1]*2)
        gm = self.model.get_gm(self.model, 0, utilities.tuplinator(ports_v), 0, self.device)
        if gm == 0:
            gm = options.gmin*2
        stamp = np.array(((gm, -gm),
                          (-gm, gm)), dtype=np.float64)
        if reduced:
            zap_rc = [pos for pos, i in enumerate(indices[1][:2]) if i == -1]
            stamp = np.delete(stamp, zap_rc, axis=0)
            stamp = np.delete(stamp, zap_rc, axis=1)
            indices = tuple(zip(*[(i, y) for i, y in zip(*indices) if (i != -1 and y != -1)]))
            stamp_flat = stamp.reshape(-1)
            stamp_folded = []
            indices_folded = []
            for ix, it in enumerate([(i, y) for i, y in zip(*indices)]):
                if it not in indices_folded:
                    indices_folded.append(it)
                    stamp_folded.append(stamp_flat[ix])
                else:
                    w = indices_folded.index(it)
                    stamp_folded[w] += stamp_flat[ix]
            indices = tuple(zip(*indices_folded))
            stamp = np.array(stamp_folded)
        return indices, stamp

    def g(self, op_index, ports_v, port_index, time=0):
        if not port_index == 0:
            raise Exception("Attepted to evaluate a custom's gm on an unknown port.")
        gm = self.model.get_gm(self.model, op_index, utilities.tuplinator(ports_v), port_index, self.device)
        return gm

    def get_op_info(self, ports_v_v):
        """Information regarding the Operating Point (OP)

        **Parameters:**

        ports_v : list of lists
            The parameter is to be set to ``[[v]]``, where ``v`` is the voltage
            applied to the custom terminals.

        **Returns:**

        op_keys : list of strings
            The labels corresponding to the numeric values in ``op_info``.
        op_info : list of floats
            The values corresponding to ``op_keys``.
        """
        vn1n2 = float(ports_v_v[0][0])
        icustom = self.i(0, (vn1n2,))
        gmcustom = self.g(0, (vn1n2,), 0)
        op_keys = ["Part ID", "V(n1-n2) [V]", "I(n1-n2) [A]", "P [W]",
                "gm [A/V]", u"T [\u00b0K]"]
        op_info = [self.part_id.upper(), vn1n2, icustom, vn1n2*icustom, gmcustom,
                   self._get_T()]
        return op_keys, op_info

    def get_netlist_elem_line(self, nodes_dict):
        ext_n1, ext_n2 = nodes_dict[self.n1], nodes_dict[self.n2]
        ret = "%s %s %s %s" % (self.part_id, ext_n1, ext_n2, self.model.name)
        # append the optional part:
        # [<AREA=float> <T=float> <IC=float> <OFF=boolean>]
        ret += " AREA=%g" % self.device.AREA
        if self.device.T is not None:
            ret += " T=%g" % self.device.T
        if self.ic is not None:
            ret += " IC=%g" % self.ic
        if self.off:
            ret += " OFF=1"
        return ret


IS_DEFAULT = 1e-14  # A
N_DEFAULT = 1.0
ISR_DEFAULT = 0.0  # A
NR_DEFAULT = 2.0
RS_DEFAULT = 0.0  # ohm
CJ0_DEFAULT = 0.0
M_DEFAULT = .5
VJ_DEFAULT = .7
FC_DEFAULT = .5
CP_DEFAULT = .0
TT_DEFAULT = .0
BV_DEFAULT = float('inf')
IBV_DEFAULT = 1e-3
KF_DEFAULT = .0
AF_DEFAULT = 1.
FFE_DEFAULT = 1.
TEMP_DEFAULT = 26.85
XTI_DEFAULT = 3.0
EG_DEFAULT = 1.11
TBV_DEFAULT = 0.0
TRS_DEFAULT = 0.0
TTT1_DEFAULT = 0.0
TTT2_DEFAULT = 0.0
TM1_DEFAULT = 0.0
TM2_DEFAULT = 0.0
T_DEFAULT = utilities.Celsius2Kelvin(26.85)
AREA_DEFAULT = 1.0

class custom_model(object):
    """A custom model implementing the `Shockley custom equation
    <http://en.wikipedia.org/wiki/Shockley_custom_equation#Shockley_custom_equation>`__.

    Currently the capacitance modeling part is missing.

    The principal parameters are:

    +---------------+-------------------+-----------------------------------+
    | *Parameter*   | *Default value*   | *Description*                     |
    +===============+===================+===================================+
    | IS            | 1e-14 A           | Specific current                  |
    +---------------+-------------------+-----------------------------------+
    | N             | 1.0               | Emission coefficient              |
    +---------------+-------------------+-----------------------------------+
    | ISR           | 0.0 A             | Recombination current             |
    +---------------+-------------------+-----------------------------------+
    | NR            | 2.0               | Recombination coefficient         |
    +---------------+-------------------+-----------------------------------+
    | RS            | 0.0 ohm           | Series resistance per unit area   |
    +---------------+-------------------+-----------------------------------+

    please refer to a textbook description of the Shockley custom equation
    or to the source file ``custom.py`` file for the other parameters.

    """
    def __init__(self, name, IS=None, N=None, ISR=None, NR=None, RS=None,
                 CJ0=None, M=None, VJ=None, FC=None, CP=None, TT=None,
                 BV=None, IBV=None, KF=None, AF=None, FFE=None, TEMP=None,
                 XTI=None, EG=None, TBV=None, TRS=None, TTT1=None, TTT2=None,
                 TM1=None, TM2=None):
        self.name = name
        self.IS = float(IS) if IS is not None else IS_DEFAULT
        self.N = float(N) if N is not None else N_DEFAULT
        self.ISR = float(ISR) if ISR is not None else ISR_DEFAULT
        self.NR = float(NR) if NR is not None else NR_DEFAULT
        self.RS = float(RS) if RS is not None else RS_DEFAULT
        self.CJ0 = float(CJ0) if CJ0 is not None else CJ0_DEFAULT
        self.M = float(M) if M is not None else M_DEFAULT
        self.VJ = float(VJ) if VJ is not None else VJ_DEFAULT
        self.FC = float(FC) if FC is not None else FC_DEFAULT
        self.CP = float(CP) if CP is not None else CP_DEFAULT
        self.TT = float(TT) if TT is not None else TT_DEFAULT
        self.BV = float(BV) if BV is not None else BV_DEFAULT
        self.IBV = float(IBV) if IBV is not None else IBV_DEFAULT
        self.KF = float(KF) if KF is not None else KF_DEFAULT
        self.AF = float(AF) if AF is not None else AF_DEFAULT
        self.FFE = float(FFE) if FFE is not None else FFE_DEFAULT
        self.TEMP = utilities.Celsius2Kelvin(
            float(TEMP)) if TEMP is not None else TEMP_DEFAULT
        self.XTI = float(XTI) if XTI is not None else XTI_DEFAULT
        self.EG = float(EG) if EG is not None else EG_DEFAULT
        self.TBV = float(TBV) if TBV is not None else TBV_DEFAULT
        self.TRS = float(TRS) if TRS is not None else TRS_DEFAULT
        self.TTT1 = float(TTT1) if TTT1 is not None else TTT1_DEFAULT
        self.TTT2 = float(TTT2) if TTT2 is not None else TTT2_DEFAULT
        self.TM1 = float(TM1) if TM1 is not None else TM1_DEFAULT
        self.TM2 = float(TM2) if TM2 is not None else TM2_DEFAULT
        self.T = T_DEFAULT
        self.last_vd = None
        self.VT = constants.Vth(self.T)

    def print_model(self):
        strm = ".model custom %s IS=%g N=%g ISR=%g NR=%g RS=%g CJ0=%g M=%g " + \
               "VJ=%g FC=%g CP=%g TT=%g BV=%g IBV=%g KF=%g AF=%g FFE=%g " + \
               "TEMP=%g XTI=%g EG=%g TBV=%g TRS=%g TTT1=%g TTT2=%g TM1=%g " + \
               "TM2=%g"
        print(strm % (self.name, self.IS, self.N, self.ISR, self.NR, self.RS,
                      self.CJ0, self.M, self.VJ, self.FC, self.CP, self.TT,
                      self.BV, self.IBV, self.KF, self.AF, self.FFE, self.TEMP,
                      self.XTI, self.EG, self.TBV, self.TRS, self.TTT1,
                      self.TTT2, self. TM1, self. TM2))

    @utilities.memoize
    def get_i(self, vext, dev):
        if dev.T != self.T:
            self.set_temperature(dev.T)
        if not self.RS:
            i = self._get_i(vext) * dev.AREA
            dev.last_vd = vext
        else:
            vd = dev.last_vd if dev.last_vd is not None else 10*self.VT
            vd = newton(self._obj_irs, vd, fprime=self._obj_irs_prime,
                        args=(vext, dev), tol=options.vea, maxiter=500)
            i = self._get_i(vext-vd)
            dev.last_vd = vd
        return i

    def _obj_irs(self, x, vext, dev):
        # obj fn for newton
        return x/self.RS-self._get_i(vext-x)*dev.AREA

    def _obj_irs_prime(self, x, vext, dev):
        # obj fn derivative for newton
        # first term
        ret = 1./self.RS
        # disable RS
        RSSAVE = self.RS
        self.RS = 0
        # second term
        ret += self.get_gm(self, 0, (vext-x,), 0, dev)
        # renable RS
        self.RS = RSSAVE
        return ret

    def _safe_exp(self, x):
        return np.exp(x) if x < 70 else np.exp(70) + 10 * x

    def _get_i(self, v):
        i = self.IS * (self._safe_exp(v/(self.N * self.VT)) - 1) \
            + self.ISR * (self._safe_exp(v/(self.NR * self.VT)) - 1)
        return i

    @utilities.memoize
    def get_gm(self, op_index, ports_v, port_index, dev):
        if dev.T != self.T:
            self.set_temperature(dev.T)
        gm = self.IS / (self.N * self.VT) *\
            self._safe_exp(ports_v[0] / (self.N * self.VT)) +\
            self.ISR / (self.NR * self.VT) *\
            self._safe_exp(ports_v[0] / (self.NR * self.VT))
        if self.RS != 0.0:
            gm = 1. / (self.RS + 1. / (gm + 1e-3*options.gmin))
        return dev.AREA * gm

    def __str__(self):
        pass

    def set_temperature(self, T):
        T = float(T)
        self.EG = constants.si.Eg(T)
        self.IS = self.IS*(T/self.T)**(self.XTI/self.N)* \
                  np.exp(-constants.e*constants.si.Eg(300)/\
                         (self.N*constants.k*T)*
                         (1 - T/self.T))
        self.BV = self.BV - self.TBV*(T - self.T)
        self.RS = self.RS*(1 + self.TRS*(T - self.T))
        self.T = T

